import click
from .kmer_count import preproc
from .motif_discovery import  scan_motif, ex_hamball, draw_logo
from .visualization import visualize_kmers

@click.group()
def cli():
    """
    KMAP: visualize kmers in 2d.
    """
    pass

def display_paper_info():
    print()
    print("This software is affiliated with the following paper:")
    print("Title: Your Paper Title")
    print("Authors: First Author, Second Author, Third Author")
    print("Journal: Journal Name")
    print("Year: 2023")
    print("DOI: https://doi.org/your-paper-doi")


cli.add_command(preproc)
cli.add_command(scan_motif)
cli.add_command(ex_hamball)
cli.add_command(draw_logo)
cli.add_command(visualize_kmers)



import bisect

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from .taichi_core import cal_ld_prob_mat_kernel, cal_cross_entropy_kernel, cal_gradient_loss_kernel, knn_smooth_kernel
import taichi as ti

from typing import List
from .kmer_count import FileNameDict
from .motif_discovery import write_lines
from pathlib import Path
import tomllib
import pickle
import click


@click.command(name="visualize_kmers")
@click.option(
    '--res_dir',
    type=str,
    help='Result directory for storing all outputs',
    required=True
    )
@click.option(
    '--debug',
    type=bool,
    default=False,
    help='display debug information.',
    required=False
    )
def visualize_kmers(res_dir: str, debug=False):
    _visualize_kmers(res_dir, debug)


def _visualize_kmers(res_dir: str, debug=False):
    # load config file
    config_file_name = FileNameDict["config_file"]  # config.toml
    config_file_path = Path(res_dir) / config_file_name
    assert config_file_path.exists()
    with open(config_file_path, "rb") as fh:
        config_dict = tomllib.load(fh)

    if not debug:
        debug = config_dict["general"]["debug"]
    n_neighbour = config_dict["visualization"]["n_neighbour"]
    random_seed = config_dict["visualization"]["random_seed"]
    n_max_iter = config_dict["visualization"]["n_max_iter"]
    learning_rate = config_dict["visualization"]["learning_rate"]
    n_best_result = config_dict["visualization"]["n_best_result"]

    if random_seed == "default":
        random_seed = None
    else:
        assert isinstance(random_seed, (int, float))

    sample_kmer_hamdist_mat_file = Path(res_dir) / FileNameDict["sample_kmer_hamdist_mat_file"]
    with open(sample_kmer_hamdist_mat_file, "rb") as fh:
        kmer_len, hamdist_mat, label_arr = pickle.load(fh)

    # dimension reduction
    ld_data = kmap(hamdist_mat, kmer_len, n_neighbour = n_neighbour,
                    n_max_iter = n_max_iter, learning_rate = learning_rate,
                    n_best_result = n_best_result, random_seed = random_seed, debug = debug)
    ld_data_lines = [f"x\ty\tlabel"]
    for x, y, label in zip(ld_data[0], ld_data[1], label_arr):
        ld_data_lines.append(f"{x:3.3f}\t{y:3.3f}\t{int(label)}")
    #for i, label in enumerate(label_arr):
    #    ld_data_lines.append(f"{ld_data[0][i]:3.3f}\t{ld_data[0][1]:3.3f}\t{int(label)}")

    ld_data_file = Path(res_dir) / FileNameDict["ld_data_file"]
    write_lines(ld_data_lines, ld_data_file)
    print("Dimensionality reduction finished. Low dimensional embeddings generated.")

    # plot 2d
    gen_fig_flag = config_dict["visualization"]["gen_fig_flag"]
    if not gen_fig_flag:
        return

    final_conseq_file = Path(res_dir) / FileNameDict["final_conseq_file"]
    assert final_conseq_file.exists()
    with open(final_conseq_file, "r") as fh:
        conseq_list = fh.read().splitlines()

    ld_fig_file_stem = Path(res_dir) / FileNameDict["ld_fig_file_stem"]
    plot_2d_data(ld_data, label_arr, conseq_list,
                       point_size=0.5, point_alpha=0.5, point_color="gray", output_fig_file_stem=str(ld_fig_file_stem))


def knn_smooth(dist_mat: np.ndarray, n_neighbour: int) -> np.ndarray:
    """
    perform smoothing to the distance matrix by sampling neighbours and take the mean
    Args:
        dist_mat: n x n matrix, each row/column is a point
        n_neighbour: number of neighbors to be used for each point
    Returns:
        smoothed distance matrix (dtype=float32)
    """
    n = len(dist_mat)
    neighbor_inds_mat = np.argpartition(dist_mat, n_neighbour, axis=1)[:, :n_neighbour]

    dist_mat = dist_mat.astype("float32")
    ret_mat = np.zeros_like(dist_mat)
    neighbor_inds_mat = neighbor_inds_mat.astype("int32")
    knn_smooth_kernel(dist_mat, neighbor_inds_mat, ret_mat, n, n_neighbour)

    ret_mat += np.transpose(ret_mat)

    return ret_mat

def gradient_loss(hd_prob_mat, ld_prob_mat, ld_data, debug=False):
    ld_data_diff = np.expand_dims(ld_data, 1) - np.expand_dims(ld_data, 0)  # n x n x 2
    prob_mat_diff = hd_prob_mat - ld_prob_mat # n x n
    ld_ratio_mat = ld_prob_mat / (1 - ld_prob_mat) # n x n
    tmp_diff_mat = ld_ratio_mat * prob_mat_diff # n x n

    assert not np.any(np.isnan(tmp_diff_mat))

    n = len(hd_prob_mat)
    ret_mat = np.zeros((n, 2))
    for i in range(n):
        for j in range(n):
            if i == j:
                continue
            ret_mat[i, :] += tmp_diff_mat[i, j] * ld_data_diff[i, j, :]
        if debug:
            print(f"{i= } {ret_mat[i, :]= }")
    return 4.0 * ret_mat


def gradient_loss_taichi(hd_prob_mat, ld_prob_mat, ld_data, debug=False):
    prob_mat_diff = hd_prob_mat - ld_prob_mat # n x n
    ld_ratio_mat = ld_prob_mat / (1 - ld_prob_mat) # n x n
    tmp_diff_mat = ld_ratio_mat * prob_mat_diff # n x n

    assert not np.any(np.isnan(tmp_diff_mat))
    assert tmp_diff_mat.dtype == np.float32
    assert ld_data.dtype == np.float32
    assert len(ld_data) == 2

    n = len(hd_prob_mat)
    ret_mat = np.zeros((2, n), dtype="float32")
    cal_gradient_loss_kernel(tmp_diff_mat, ld_data, ret_mat, n)

    return 4.0 * ret_mat


def cross_entropy(hd_prob_mat, ld_prob_mat) -> float:
    """
    calculate the cross entropy between hd_prob_mat and ld_prob_mat
    Args:
        hd_prob_mat: high dimensional data probability matrix
        ld_prob_mat: low dimensional data probability matrix
    Returns:
        cross entropy, one float number
    """
    tmpmat = - hd_prob_mat * np.log(ld_prob_mat) - (1 - hd_prob_mat) * np.log(1 - ld_prob_mat)
    np.fill_diagonal(tmpmat, 0.0)
    return np.sum(tmpmat)


def cross_entropy_taichi(hd_prob_mat, ld_prob_mat, iter_mat) -> float:
    """
    calculate the cross entropy between hd_prob_mat and ld_prob_mat
    Args:
        hd_prob_mat: high dimensional data probability matrix
        ld_prob_mat: low dimensional data probability matrix
    Returns:
        cross entropy, one float number
    """
    assert hd_prob_mat.dtype == np.float32
    assert ld_prob_mat.dtype == np.float32
    cross_entropy_mat = np.zeros_like(hd_prob_mat)
    i_ind_arr, j_ind_arr = iter_mat[0], iter_mat[1]
    cal_cross_entropy_kernel(hd_prob_mat, ld_prob_mat, len(i_ind_arr), i_ind_arr, j_ind_arr, cross_entropy_mat)
    return np.sum(cross_entropy_mat) * 2


def add_jitter(ld_data: np.ndarray, eps: float) -> np.ndarray:
    """
    add jitters to identical points
    Args:
        ld_data: low dimensional data
        eps: cutoff of identical points
    Returns:
        updated low dimensional data
    """
    def _add_jitter_1d(ld_data, d: int):
        idx = np.argsort(ld_data[:, d])
        sort_arr = ld_data[idx, d]
        tmpinds = np.where(np.diff(sort_arr) < eps)[0]
        ld_data[idx[tmpinds], d] += np.random.normal(0, 0.01, len(tmpinds))
        return ld_data
    ld_data = _add_jitter_1d(ld_data, 0)
    ld_data = _add_jitter_1d(ld_data, 1)
    return ld_data


def sigmoid(dist_mat: np.ndarray, max_val=16.0, change_point=10.0, scale_factor=3.0) -> np.ndarray:
    """
    transform the distance matrix
    Args:
        dist_mat: distance matrix
        max_val: maximum value of the distance
        change_point: change point of the distance distribution
        scale_factor: scale factor for the conversion
    Returns:
        converted distance matrix
    """
    assert max_val > change_point > 0
    assert scale_factor > 0
    return max_val / (1 + np.exp(-scale_factor * (dist_mat - change_point)))


def cal_euclidean_dist2_mat(ld_data: np.ndarray) -> np.ndarray:
    """
    calculate the euclidean distance square for each pair of input data
    Args:
        ld_data: low dimensional data, n x 2 matrix
    Returns:
        Euclidean distance matrix, n x n
    """
    a = ld_data
    b = a.reshape(a.shape[0], 1, a.shape[1])

    return np.einsum('ijk, ijk->ij', a - b, a - b)


def cal_ld_prob_mat(ld_dist2_mat: np.ndarray):
    q = 1.0 / (1 + ld_dist2_mat)
    q = np.minimum(1 - 1e-12, q)
    return q


def cal_ld_prob_mat_taichi(ld_data: np.ndarray, iter_mat: np.ndarray):
    """
    calculate the probability matrix for the low dimensional data
    Args:
        ld_data: 2 x n matrix
        iter_mat: iteration index (2 x n matrix) for i,j pairs such that i<j, iter_mat[0] are index for i, iter_mat[1] are index for j
    Returns:
        n x n probability matrix
    """
    assert ld_data.shape[0] == 2
    assert ld_data.dtype == np.float32
    assert iter_mat.dtype == np.int32
    n = len(ld_data[0])

    x_arr, y_arr = ld_data[0], ld_data[1]
    i_ind_arr, j_ind_arr = iter_mat[0], iter_mat[1]
    prob_mat = np.ones((n, n), dtype="float32")
    cal_ld_prob_mat_kernel(x_arr, y_arr, len(i_ind_arr), i_ind_arr, j_ind_arr, prob_mat)
    eps = 1e-3
    prob_mat = np.minimum(prob_mat, 1 - eps)
    prob_mat = np.maximum(prob_mat, eps)
    return prob_mat


def kmap(hamdist_mat: np.ndarray, kmer_len: int, n_neighbour=20,
         n_max_iter=2500, learning_rate=0.01, n_best_result=10, random_seed=None, debug=True) -> np.ndarray:
    trans_dist_mat = knn_smooth(dist_mat=hamdist_mat, n_neighbour=n_neighbour)
    trans_dist_mat = sigmoid(trans_dist_mat, 16.0, change_point=kmer_len/2, scale_factor=0.2 * kmer_len - 0.2)
    print("distance smoothing finished.")
    ld_data = umap(trans_dist_mat, n_max_iter=n_max_iter, learning_rate=learning_rate,
                   n_best_result=n_best_result, random_seed=random_seed, debug=debug)
    print("optimization finished.")
    return  ld_data


def umap(hd_dist_mat: np.ndarray, n_max_iter=2500, learning_rate=0.01, n_best_result=10, random_seed=None, debug=True) -> np.ndarray:
    """
    perform dimension reduction for kmer distance matrix
    Args:
        hd_dist_mat: distance matrix of kmers (high dimensional data)
        n_max_iter: maximum number of iterations in the optimization
        learning_rate: learning rate in the optimization
        n_best_result: record the 10 best results
        random_seed: random seed for the optimization, None means random seed
    Returns: low dimensional representation of kmers in the form of n x 2 matrix
    """
    np.random.seed(random_seed)

    n_data = len(hd_dist_mat)
    sigma0 = 0.5

    iter_mat = np.array([(i, j) for i in range(n_data) for j in range(n_data) if i < j]).T.astype("int32")
    iter_mat = np.ascontiguousarray(iter_mat)

    hd_prob_mat = np.exp(-hd_dist_mat / sigma0).astype("float32")

    # init low dimensional representation
    ld_data = np.random.randn(2, n_data).astype("float32")
    best_res_list = [(np.array([np.inf]), np.random.randn(2, n_data).astype("float32")) for _ in range(n_best_result)]

    loss = np.inf
    for i_iter in range(n_max_iter):
        if debug:
            print(f"{i_iter= } {loss= }")

        ld_prob_mat = cal_ld_prob_mat_taichi(ld_data, iter_mat)
        curr_loss = cross_entropy_taichi(hd_prob_mat, ld_prob_mat, iter_mat)

        if curr_loss < best_res_list[-1][0]:
            tmpres = best_res_list[-1]
            tmpres[0][0] = curr_loss
            tmpres[1][:] = ld_data[:]
            best_res_list = best_res_list[:-1]
            bisect.insort_right(best_res_list, tmpres, key=lambda x: x[0][0])

        if abs(loss - curr_loss) < 1e-7 * abs(curr_loss):
            break

        loss = curr_loss
        grad_loss = gradient_loss_taichi(hd_prob_mat, ld_prob_mat, ld_data)  # O(n^2)

        ld_data += (-grad_loss * learning_rate)
        ld_data = add_jitter(ld_data, eps=0.1) # nlog(n)

    if debug:
        print()
        print("Best results")
        for i in range(n_best_result):
            print(f"{i= } loss={best_res_list[i][0][0]:0.2f}")

    ld_data = best_res_list[0][1]
    return ld_data


def plot_2d_data(ld_data: np.ndarray, label_arr: np.ndarray=None, conseq_list: List[str]=[], cmap: str = "Dark2",
                 point_size=0.5, point_alpha=0.5, point_color="gray", output_fig_file_stem: str|Path = None):
    # cmap: 'Pastel1', 'Pastel2', 'Paired', 'Accent', 'Dark2',
    #                       'Set1', 'Set2', 'Set3', 'tab10', 'tab20', 'tab20b',
    #                       'tab20c'
    assert ld_data.shape[0] == 2
    assert len(ld_data[0]) == len(label_arr)
    x_arr, y_arr = ld_data[0], ld_data[1]

    if cmap=="Dark2":
        cmap = ListedColormap(plt.get_cmap("Dark2").colors[:7]) # remove the last gray color

    # create the figure
    fig, ax = plt.subplots(figsize=(15, 15))  # 15 inch

    if label_arr is None:
        plt.scatter(x_arr, y_arr, s=point_size, c=point_color)
        if output_fig_file_stem:
            plt.savefig(output_fig_file_stem + ".png", format="png")
            plt.savefig(output_fig_file_stem + ".pdf", format="pdf")
        plt.show()
        return

    max_label = max(label_arr)
    random_kmer_inds = label_arr == max_label
    motif_kmer_inds = np.logical_not(random_kmer_inds)

    if len(conseq_list) == 0:
        conseq_list = [f"motif-{i}" for i in range(max_label)]
    else:
        assert len(conseq_list) == max_label
        conseq_list = [f"{conseq_list[i]}" for i in range(max_label)]

    ax.scatter(x_arr[random_kmer_inds], y_arr[random_kmer_inds],
                s=point_size, c=point_color, alpha=point_alpha)
    scatter = ax.scatter(x_arr[motif_kmer_inds], y_arr[motif_kmer_inds],
                s=point_size*1.2, c=label_arr[motif_kmer_inds], alpha=0.9, cmap=cmap)

    handles, labels = scatter.legend_elements()
    ax.legend(handles, conseq_list, loc="upper right", title="motifs")

    if output_fig_file_stem:
        plt.savefig(output_fig_file_stem + ".png", format="png")
        plt.savefig(output_fig_file_stem + ".pdf", format="pdf")

    plt.show()



if __name__=="__main__":

    GPU_MODE = False
    ti.init(arch=ti.cuda, default_ip=ti.i64)
    if ti.cfg.arch == ti.cuda:
        GPU_MODE = True
        print("GPU is available")
    else:
        GPU_MODE = False
        print("GPU is not available")

    hd_dim = 3
    hd_sigma = 50
    hd_data_0 = np.random.normal(0.0, 10, (100, hd_dim)) + np.random.normal(0, hd_sigma, hd_dim)
    hd_data_1 = np.random.normal(0.0, 10, (100, hd_dim)) + np.random.normal(0, hd_sigma, hd_dim)
    hd_data_2 = np.random.uniform(-2.0 * hd_sigma, 2.0 * hd_sigma, (100, hd_dim))
    hd_data = np.vstack((hd_data_0, hd_data_1, hd_data_2))
    print(hd_data.shape)

    hd_dist_mat = np.sqrt(cal_euclidean_dist2_mat(hd_data))
    print(hd_dist_mat.shape)

    #plt.scatter(hd_data[:, 0], hd_data[:, 1])
    #plt.show()

    ld_data = kmap(hd_dist_mat, 2*hd_sigma, 3 * 10, n_max_iter=50).T
    inds = np.arange(200)
    plt.scatter(ld_data[inds, 0], ld_data[inds, 1], color="r")
    inds = np.arange(200, 300)
    plt.scatter(ld_data[inds, 0], ld_data[inds, 1], color="b")
    plt.show()














from .inimotif import *
from src.kmap.kmer_count import *
from src.kmap.motif_discovery import *
from src.kmap.visualization import kmap, plot_2d_data, knn_smooth, _visualize_kmers
from src.kmap.motif_discovery import _convert_to_block_arr, _convert_to_block_mat, _scan_motif
from src.kmap.kmer_count import _preproc


import random

# TODO: TEST and parameter modification
def main(input_fasta_file, min_kmer_len, max_kmer_len,
         mask_consensus_list: List[str]=(),
         mask_noise_seq_list: List[str]=(),
         motif_def_file='.src/kmap/default_motif_def_table.csv',
         out_dir=".", debug=False):
    def write_csv(res, outfile):
        with open(outfile, 'w+') as fh:
            for line in res:
                fh.write(",".join(str(line)))

    motif_def_dict = init_motif_def_dict(motif_def_file)

    # preprocess: convert input fasta file into pkl file
    proc_input(input_fasta_file, out_dir)

    # load input pickle file as numpy array
    input_pkl_file = out_dir + f"/input.bin.pkl"
    with open(input_pkl_file, "rb") as fh:
        seq_np_arr = pickle.load(fh)

    # mask user provided noise sequnces from the input sequence
    if len(mask_noise_seq_list) > 0:
        max_ham_dist_list = [0 for _ in mask_noise_seq_list]
        seq_np_arr = mask_ham_ball(seq_np_arr, motif_def_dict, mask_noise_seq_list, max_ham_dist_list)

    # mask user provided consensus (Hamming ball) from the input sequence
    if len(mask_consensus_list) > 0:
        seq_np_arr = mask_ham_ball(seq_np_arr, motif_def_dict, mask_consensus_list)

    orig_seq_np_arr = seq_np_arr.copy()

    res = ["kmer_len,hash,kmer,proportion"]
    for kmer_len in range(min_kmer_len, max_kmer_len):
        seq_np_arr[:] = orig_seq_np_arr[:]
        p_uniform_k = motif_def_dict[kmer_len].p_uniform
        max_ham_dist = motif_def_dict[kmer_len].max_ham_dist
        p_cut_off = 1.3
        top_k = 5
        n_trial = 10
        revcom_mode = True

        assert min_kmer_len <= kmer_len <= max_kmer_len
        kmer_cnt_file = out_dir + f"/k{kmer_len}.pkl" # [kmer_len, uniq_kh_arr, uniq_kh_cnt_arr]
        consensus_kh_dict = find_motif(seq_np_arr, kmer_len, max_ham_dist, p_uniform_k, p_cut_off, top_k, n_trial,
                                       revcom_mode, save_kmer_cnt_flag=True,
                                       kmer_cnt_pkl_file=kmer_cnt_file, debug=debug)  # snp_np_arr are mutated in find_motif()

        print(f"filtered consensus kmers when k = {kmer_len}")
        for kh, prop in consensus_kh_dict.items():
            print(f'{hash2kmer(kh, kmer_len)} perc={prop * 100:0.5f}%')
            res.append(f"{kmer_len},{kh},{hash2kmer(kh, kmer_len)},{prop:0.8f}")

    outfile = out_dir + "/res_kmers.csv"
    write_csv(res, outfile)


def test_init_motif_def_dict():
    motif_def_dict = init_motif_def_dict("../src/kmap/default_motif_def_table.csv")
    print(motif_def_dict)
    assert len(motif_def_dict) == 18
    assert isinstance(motif_def_dict[7], MotifDef)


def gen_test_fa_file(conseq1="AATCGATAGC", conseq2="AGGACCTACGTAC", n_seq=1000, min_len=30, max_len=60, output_seq_file='./test/test.fa'):
    import random

    def generate_random_sequence(length):
        bases = ['A', 'C', 'G', 'T']
        return ''.join(random.choice(bases) for _ in range(length))

    def gen_motif_seq(conseq, length, mutation_rate=0.05):
        # 54% sequence are expected to be the same as conseq when mutation rate is 0.05, (1-0.05)**12=0.54
        conseq_len = len(conseq)
        bases = ['A', 'C', 'G', 'T']
        seq_base_list = [random.choice(bases) for _ in range(length)]
        pos = random.randint(0, length - conseq_len)
        for i, b in enumerate(conseq):
            if random.random() > mutation_rate:
                seq_base_list[pos + i] = b
        return "".join(seq_base_list)

    n_seq_1 = int(n_seq * 0.4)
    n_seq_2 = int(n_seq * 0.4)
    n_seq_other = n_seq - n_seq_1 - n_seq_2

    motif1_list = [gen_motif_seq(conseq1, random.randint(min_len, max_len)) for _ in range(n_seq_1)]
    motif2_list = [gen_motif_seq(conseq2, random.randint(min_len, max_len)) for _ in range(n_seq_2)]
    rand_list = [generate_random_sequence(random.randint(min_len, max_len)) for _ in range(n_seq_other)]

    with open(output_seq_file, 'w+') as fh:
        k = 0
        for s in motif1_list:
            fh.write(f'>seq{k}_1\n')
            fh.write(s + '\n')
            k += 1
        for s in motif2_list:
            fh.write(f'>seq{k}_2\n')
            fh.write(s + '\n')
            k += 1
        for s in rand_list:
            fh.write(f'>seq{k}_rand\n')
            fh.write(s + '\n')
            k += 1


def test_gen_rand_fa_file(n_seq=100, min_len=30, max_len=60, rand_seq_file='random_sequences.fasta'):
    import random

    def generate_random_sequence(length):
        bases = ['A', 'C', 'G', 'T']
        return ''.join(random.choice(bases) for _ in range(length))

    def generate_random_fasta_file(file_path, num_sequences):
        with open(file_path, 'w') as fasta_file:
            for i in range(num_sequences):
                sequence_length = random.randint(min_len, max_len)
                sequence = generate_random_sequence(sequence_length)
                fasta_file.write(f'>seq{i}\n')
                fasta_file.write(sequence + '\n')

    generate_random_fasta_file(rand_seq_file, n_seq)


def test_proc_input():
    input_fasta_file = "./tests/test.fa"
    res_dir = "./test"

    proc_input(input_fasta_file, res_dir, out_bin_file_name = "input.bin.pkl", debug = True)
    #rand_seq_file = "test_rand_seq.fasta"
    #test_gen_rand_fa_file(rand_seq_file=rand_seq_file)
    #proc_input("test_rand_seq.fasta", out_dir ="..", debug = True)


def test_preproc():
    input_fasta_file = "./tests/test.fa"
    res_dir = "./test"
    config_dict, motif_def_dict = _preproc(input_fasta_file, res_dir)
    print(config_dict)
    print(motif_def_dict)


def test_read_default_config_file():
    config_dict = read_default_config_file(debug=True)
    motid_def_dict = gen_motif_def_dict(config_dict, debug=True)
    print(config_dict)


def test_comp_kmer_hash_taichi() -> Tuple:
    kmer_len = 10
    input_binary_file = "../input.bin.pkl"
    with open(input_binary_file, "rb") as fh:
        seq_np_arr = pickle.load(fh)

    res = comp_kmer_hash_taichi(seq_np_arr, kmer_len)
    uniq_kh_arr, uniq_kh_cnt_arr = count_uniq_hash(res, kmer_len)
    print(res)
    print(uniq_kh_arr, uniq_kh_cnt_arr)


def test_count_uniq_hash():
    seq = "TTTTCGTNCACGACGCTACCTTAAAGCATCCTTCTNTGATACCATAGANNNNNGCAGCTCCTTATCGTTTTAGCTTTCGTATTCGTCTAATCGTCTTTTACTCGACGAAAA"
    kmer_len = 3
    from tests.inimotif import KmerCounter
    kc8 = KmerCounter(kmer_len, revcom_flag=False, unique_kmer_in_seq_mode=False)
    kmer_dict = kc8.scan_seq(seq)
    c8 = Counter(kmer_dict)

    seq_np_arr = dna2arr(seq)
    hash_arr = comp_kmer_hash_taichi(seq_np_arr, kmer_len)
    res = count_uniq_hash(hash_arr, kmer_len)
    c8_new = Counter(dict(zip(res[0], res[1])))

    assert c8 == c8_new

    print(c8)
    print(c8_new)


def test_mask_input():
    seq = "TTTTCGTNCACGACGCTACCTTAAAGCATCCTTCTNTGATACCATAGANNNNNGCAGCTCCTTATCGTTTTAGCTTTCGTATTCGTCTAATCGTCTTTTACTCGACGAAAA"
    kmer_len = 5
    from tests.inimotif import KmerCounter
    kc8 = KmerCounter(kmer_len, revcom_flag=False, unique_kmer_in_seq_mode=False)

    consensus_kh = kc8.kmer2hash(seq[0:5])
    max_hamball_dist = 2

    seq_np_arr = dna2arr(seq)
    seq_np_arr1 = mask_input(seq_np_arr, kmer_len, np.array([consensus_kh]), np.array([max_hamball_dist]))
    print(seq_np_arr1)

    seq_np_arr1[seq_np_arr1==MISSING_VAL] = 4
    alphabet = {0: 'A', 1: 'C', 2: 'G', 3: 'T', 4: 'N'}
    seq_arr = [alphabet[e] for e in seq_np_arr1]
    print("".join(seq_arr))
    print(seq)


def test_merge_revcom():
    kmer_len = 3
    kh_arr = np.array([0, 2, 10, 11, 17, 18, 19, 23, 27, 33, 36, 38, 41, 43, 46, 51, 53, 57, 59])
    rc_kh_arr = np.array([revcom_hash(i, kmer_len) for i in kh_arr])
    cnt_arr = np.ones_like(kh_arr)

    comm_kh, comm_kh_nat_inds, comm_kh_rc_inds = np.intersect1d(kh_arr, rc_kh_arr, return_indices=True)
    print(f"{comm_kh= }")
    print(f"{comm_kh_nat_inds}")
    print(f"{comm_kh_rc_inds}")
    assert np.all(np.sort(comm_kh_nat_inds) == np.sort(comm_kh_rc_inds))

    cnt_before_merge = sum(cnt_arr)
    merge_kh_arr, merge_kh_cnt_arr = merge_revcom(kh_arr, cnt_arr, kmer_len, keep_lower_hash_flag=True)
    assert all(merge_kh_cnt_arr < 32)
    assert cnt_before_merge == sum(merge_kh_cnt_arr)

    for kh in [10, 17, 36]:
        assert merge_kh_cnt_arr[merge_kh_arr == kh] == 2

    kmer_len = 4
    kh_arr = np.random.choice(256, 1000)
    uniq_kh_arr, uniq_kh_cnt_arr = np.unique(kh_arr, return_counts=True)
    cnt_before_merge = sum(uniq_kh_cnt_arr)
    merge_kh_arr, merge_kh_cnt_arr = merge_revcom(uniq_kh_arr, uniq_kh_cnt_arr, kmer_len, keep_lower_hash_flag=True)
    assert all(merge_kh_cnt_arr < 256/2)
    assert cnt_before_merge == sum(merge_kh_cnt_arr)


def test_kmer2hash():
    kmer = "ACTGA"
    kh = kmer2hash(kmer)

    kc8 = KmerCounter(len(kmer), revcom_flag=False, unique_kmer_in_seq_mode=False)
    assert kh == kc8.kmer2hash(kmer)

    kmer = "ACTACTGGAGGACCTACGTAAGCCACGA"
    kh = kmer2hash(kmer)

    kc8 = KmerCounter(len(kmer), revcom_flag=False, unique_kmer_in_seq_mode=False)
    assert kh == kc8.kmer2hash(kmer)


def test_hash2kmer():
    kmer = "ACTGA"
    kc8 = KmerCounter(len(kmer), revcom_flag=False, unique_kmer_in_seq_mode=False)

    kh = kc8.kmer2hash(kmer)
    assert hash2kmer(kh, len(kmer)) == kmer

    kmer = "ACTACTGGAGGACCTACGTAAGCCACGA"
    kc8 = KmerCounter(len(kmer), revcom_flag=False, unique_kmer_in_seq_mode=False)

    kh = kc8.kmer2hash(kmer)
    assert hash2kmer(kh, len(kmer)) == kmer

def test_mask_ham_ball():
    seqs = "AAAAAAAAAAAAAAAAAAAAAACTAGCTGCCAGTCCCCCCCCCCC"
    seq_np_arr = dna2arr(seqs)[:-1]
    motif_def_dict = init_motif_def_dict("../src/kmap/default_motif_def_table.csv")
    consensus_seq_list = ["AAA", "CCCC"]
    max_ham_dist_list = [0, 0]
    res = mask_ham_ball(seq_np_arr, motif_def_dict, consensus_seq_list, max_ham_dist_list)
    res_str = arr2dna(res)
    assert res_str == "NNNNNNNNNNNNNNNNNNNNNNCTAGCTGCCAGTNNNNNNNNNNN"

    seqs = "AAAAAAAAAAAAAAAAAAAAAACTAGCTGGGGGGGGGGGGGGGGGGGGGGGGGGCCAGTCCCCCCCCCCC"
    seq_np_arr = dna2arr(seqs)[:-1]
    motif_def_dict = init_motif_def_dict("../src/kmap/default_motif_def_table.csv")
    consensus_seq_list = ["AAAAAAA", "CCCCCCCC", "GGGGGGGGG"]
    res = mask_ham_ball(seq_np_arr, motif_def_dict, consensus_seq_list)
    res_str = arr2dna(res)
    assert res_str == "NNNNNNNNNNNNNNNNNNNNNNNTANNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNANNNNNNNNNNNNN"


def test_find_motif():
    input_pkl_file = f"../input.bin.pkl"
    with open(input_pkl_file, "rb") as fh:
        seq_np_arr = pickle.load(fh)
    kmer_len = 8
    motif_def_dict = init_motif_def_dict("../src/kmap/default_motif_def_table.csv")
    max_ham_dist = motif_def_dict[kmer_len].max_ham_dist
    ratio_mu = motif_def_dict[kmer_len].ratio_mu
    ratio_std = motif_def_dict[kmer_len].ratio_std
    ratio_cutoff = motif_def_dict[kmer_len].ratio_cutoff
    p_unif = motif_def_dict[kmer_len].p_uniform
    kmer_cnt_pkl_file = f"./k{kmer_len}.pkl"

    res_dict = find_motif(seq_np_arr, kmer_len, max_ham_dist, p_unif,
                          ratio_mu, ratio_std, ratio_cutoff, kmer_cnt_pkl_file=kmer_cnt_pkl_file)
    for kh, values in res_dict.items():
        kh_prop, kh_ratio_log10_pvalue = values
        print(f'{hash2kmer(kh, kmer_len)} {kh_prop*100:02f}% {kh_ratio_log10_pvalue:0.4f}')


def test_sample_disp_kmer():
    consensus_seq_list = ["AACCCTAACCCT", "GGAAGGAA"]
    kmer_len = 12
    motif_def_dict = init_motif_def_dict("../src/kmap/default_motif_def_table.csv")
    out_dir = "../"
    sample_kmer_kh, sample_kmer_count, sample_kmer_label, consensus_list = \
        sample_disp_kmer(consensus_seq_list, kmer_len, motif_def_dict, Path(out_dir))
    #convert hash to kmer
    sample_kmer_list = [hash2kmer(kh, kmer_len) for kh in sample_kmer_kh]
    assert len(sample_kmer_list) == len(sample_kmer_count) == len(sample_kmer_label)
    print(sum(sample_kmer_count))
    print(sum(sample_kmer_count[sample_kmer_label == 0]))
    print(sum(sample_kmer_count[sample_kmer_label == 1]))
    print(sum(sample_kmer_count[sample_kmer_label == 2]))

    print(sample_kmer_label)
    print(sample_kmer_list)

def test_sample_disp_kmer_gen_data(test_dir="./test"):
    def rand_seq(n):
        return random.choices(["A", "C", "G", "T"], k=n)

    def ham_dist(seq1, seq2, n=0):
        # assert len(seq1)==len(seq2)
        if n == 0:
            n = len(seq1)
        dist = 0
        for i in range(n):
            if seq1[i] != seq2[i]:
                dist += 1
        return dist

    def sample_data(conseq1, conseq2, n_seq):
        assert len(conseq1) < len(conseq2)
        label_list = []
        seq_list = []

        conseq = conseq1
        for i in range(n_seq):
            tmpseq = [c for c in conseq] + rand_seq(len(conseq2) - len(conseq1))
            tmp_pos = np.random.randint(0, len(conseq) - 1)
            tmpseq[tmp_pos] = random.sample(["A", "C", "G", "T"], 1)[0]
            seq_list.append("".join(tmpseq))
            label_list.append(0)

        conseq = conseq2
        for i in range(n_seq):
            tmpseq = [c for c in conseq]
            tmp_pos = np.random.randint(0, len(conseq) - 1)
            tmpseq[tmp_pos] = random.sample(["A", "C", "G", "T"], 1)[0]
            seq_list.append("".join(tmpseq))
            label_list.append(1)

        for i in range(2 * n_seq):
            tmpseq = rand_seq(len(conseq2))
            seq_list.append("".join(tmpseq))
            label_list.append(2)

        return seq_list, label_list

    conseq1 = "ACGTGAGGA"
    conseq2 = "CGGACATAGTGA"
    n_seq = 10000
    kmer_len = len(conseq2)
    overwrite_flag = True

    # create test file
    if not os.path.exists(test_dir):
        os.mkdir(test_dir)
    input_fasta_file = f"{test_dir}/test.fa"
    if not overwrite_flag and os.path.isfile(input_fasta_file):
        seq_list = []
        label_list = []
        with open(input_fasta_file, "r") as fh:
            for rec in SeqIO.parse(fh, "fasta"):
                seq_list.append(str(rec.seq))
                label_list.append(int(rec.name.split("_")[1]))
    else:
        seq_list, label_list = sample_data(conseq1, conseq2, n_seq)
        with open(input_fasta_file, "w") as fh:
            i = 0
            for seq, label in zip(seq_list, label_list):
                fh.write(f">{i}_{label}\n")
                fh.write(f"{seq}\n")
                # print(f"{seq} {label}")
                i = i + 1

    min_k = len(conseq1)
    max_k = len(conseq2) + 1
    user_consensus_list = []
    user_noise_list = []
    main(input_fasta_file, min_k, max_k, user_consensus_list, user_noise_list, out_dir=test_dir, debug=True)

    # todo: collect consensus sequences, merge them

    consensus_seq_list = [conseq1, conseq2]
    motif_def_dict = init_motif_def_dict("../src/kmap/default_motif_def_table.csv")
    out_dir = test_dir
    sample_kmer_kh, sample_kmer_count, sample_kmer_label, consensus_list = \
        sample_disp_kmer(consensus_seq_list, kmer_len, motif_def_dict, Path(out_dir))

    # should be close to n_seq, n_seq, 2*n_seq
    print(f"label0 count: {sum(sample_kmer_count[sample_kmer_label == 0])}")
    print(f"label1 count: {sum(sample_kmer_count[sample_kmer_label == 1])}")
    print(f"label2 count: {sum(sample_kmer_count[sample_kmer_label == 2])}")

    i = 0
    combined_list = sorted(zip(sample_kmer_kh, sample_kmer_count, sample_kmer_label), key=lambda x: x[2])
    for kh, cnt, label in combined_list:
        if label == len(consensus_list):
            conseq = conseq2
        else:
            conseq = consensus_seq_list[label]

        if label == 0:
            h_dist = ham_dist(hash2kmer(kh, kmer_len), conseq, len(conseq1))
        elif label == 1:
            h_dist = ham_dist(hash2kmer(kh, kmer_len), conseq, len(conseq2))
        else:
            h_dist = ham_dist(hash2kmer(kh, kmer_len), conseq, len(conseq2))

        print(f"{i= } {label=} {h_dist= }")
        i += 1

    return sample_kmer_kh, sample_kmer_count, sample_kmer_label, consensus_list


def test_convert_to_block_mat():
    samp_cnts = np.array([2, 1, 4])
    uniq_dist_mat = np.random.randint(0, 100, (3, 3))
    hamdist_mat = _convert_to_block_mat(uniq_dist_mat, samp_cnts)

    print(samp_cnts)
    print(uniq_dist_mat)
    print(hamdist_mat)


def test_cal_samp_kmer_hamdist_mat():
    def rand_seq(n):
        return random.choices(["A", "C", "G", "T"], k=n)
    def ham_dist(seq1, seq2, n=0):
        #assert len(seq1)==len(seq2)
        if n==0:
            n = len(seq1)
        dist = 0
        for i in range(n):
            if seq1[i] != seq2[i]:
                dist += 1
        return dist

    def sample_data(conseq1, conseq2, n_seq):
        assert len(conseq1) < len(conseq2)
        label_list = []
        seq_list = []

        conseq = conseq1
        for i in range(n_seq):
            tmpseq = [c for c in conseq] + rand_seq(len(conseq2) - len(conseq1))
            tmp_pos = np.random.randint(0, len(conseq) - 1)
            tmpseq[tmp_pos] = random.sample(["A", "C", "G", "T"], 1)[0]
            seq_list.append("".join(tmpseq))
            label_list.append(0)

        conseq = conseq2
        for i in range(n_seq):
            tmpseq = [c for c in conseq]
            tmp_pos = np.random.randint(0, len(conseq) - 1)
            tmpseq[tmp_pos] = random.sample(["A", "C", "G", "T"], 1)[0]
            seq_list.append("".join(tmpseq))
            label_list.append(1)

        for i in range(2 * n_seq):
            tmpseq = rand_seq(len(conseq2))
            seq_list.append("".join(tmpseq))
            label_list.append(2)

        return seq_list, label_list

    conseq1 = "ACGTGAGGA"
    conseq2 = "CGGAGAGAGTA"
    n_seq = 50
    kmer_len = len(conseq2)
    overwrite_flag = True

    # create test file
    input_fasta_file = "../test.fa"
    if not overwrite_flag and os.path.isfile(input_fasta_file):
        seq_list = []
        label_list = []
        with open(input_fasta_file, "r") as fh:
            for rec in SeqIO.parse(fh, "fasta"):
                seq_list.append(str(rec.seq))
                label_list.append(int(rec.name.split("_")[1]))
    else:
        seq_list, label_list = sample_data(conseq1, conseq2, n_seq)
        with open(input_fasta_file, "w") as fh:
            i = 0
            for seq, label in zip(seq_list, label_list):
                fh.write(f">{i}_{label}\n")
                fh.write(f"{seq}\n")
                #print(f"{seq} {label}")
                i = i + 1

    min_k = len(conseq1)
    max_k = len(conseq2)+1
    user_consensus_list = []
    user_noise_list = []
    main(input_fasta_file, min_k, max_k, user_consensus_list, user_noise_list, debug=True)

    consensus_seq_list = [conseq1, conseq2]
    motif_def_dict = init_motif_def_dict("../src/kmap/default_motif_def_table.csv")
    out_dir = "../"
    sample_kmer_kh, sample_kmer_count, sample_kmer_label, consensus_list = \
        sample_disp_kmer(consensus_seq_list, kmer_len, motif_def_dict, Path(out_dir))

    # should be close to n_seq, n_seq, 2*n_seq
    print(f"label0 count: {sum(sample_kmer_count[sample_kmer_label==0])}")
    print(f"label1 count: {sum(sample_kmer_count[sample_kmer_label == 1])}")
    print(f"label2 count: {sum(sample_kmer_count[sample_kmer_label == 2])}")

    i = 0
    combined_list = sorted(zip(sample_kmer_kh, sample_kmer_count, sample_kmer_label), key=lambda x: x[2])
    for kh, cnt, label in combined_list:
        if label == len(consensus_list):
            conseq = conseq2
        else:
            conseq = consensus_seq_list[label]

        if label==0:
            h_dist = ham_dist(hash2kmer(kh, kmer_len), conseq, len(conseq1))
        elif label==1:
            h_dist = ham_dist(hash2kmer(kh, kmer_len), conseq, len(conseq2))
        else:
            h_dist = ham_dist(hash2kmer(kh, kmer_len), conseq, len(conseq2))

        print(f"{i= } {label=} {h_dist= }")
        i += 1

    inds = np.argsort(sample_kmer_label)
    sample_kmer_kh = sample_kmer_kh[inds]
    sample_kmer_count = sample_kmer_count[inds]
    sample_kmer_label = sample_kmer_label[inds]

    for kh, label in zip(sample_kmer_kh, sample_kmer_label):
        print(f"label={label} kh={kh} seq={hash2kmer(kh, kmer_len)}")

    uniq_dist_mat = cal_samp_kmer_hamdist_mat(sample_kmer_kh, sample_kmer_count, sample_kmer_label,
                                         consensus_list, kmer_len, uniq_dist_flag=True)
    assert np.all(uniq_dist_mat.diagonal() == 0)
    assert np.all(uniq_dist_mat == np.transpose(uniq_dist_mat))

    for i, conseq in enumerate([conseq1, conseq2]):
        tmpinds = np.where(sample_kmer_label == i)[0]
        tmpinds1 = np.where(sample_kmer_label == 2)[0]
        idx = np.ix_(tmpinds, tmpinds)
        idx1 = np.ix_(tmpinds, tmpinds1)
        print(i)
        assert np.all(uniq_dist_mat[idx] <= motif_def_dict[len(conseq)].max_ham_dist*2)
        assert np.mean(uniq_dist_mat[idx]) < np.mean(uniq_dist_mat[idx1])

    print(uniq_dist_mat)


def test_gen_motif_co_occurence_mat():
    input_fasta_file = "test1.fa"
    conseq_list = ["ACGT", "GAAG"]
    motif_def_dict = init_motif_def_dict("../src/kmap/default_motif_def_table.csv")
    res = gen_motif_co_occurence_mat(input_fasta_file, conseq_list, motif_def_dict, revcom_mode = True)
    assert np.all(res == np.array([[0, 2], [2, 0]]))
    print(res)


def test_knn_smooth():
    def knn_smooth1(dist_mat: np.ndarray, n_neighbor: int) -> np.ndarray:
        """
        perform smooth to the distance matrix by sampling neighbours and take the mean
        Args:
            dist_mat: n x n matrix, each row/column is a point
            n_neighbor: number of neighbors to be sampled for each point
        Returns:
            smoothed distance matrix
        """
        n = len(dist_mat)
        neighbor_inds_mat = np.argpartition(dist_mat, n_neighbor, axis=1)[:, :n_neighbor]

        ret_mat = np.zeros_like(dist_mat, dtype=float)
        for i in range(n):
            for j in range(i + 1, n):
                tmp_sum = 0.0
                for ii in neighbor_inds_mat[i]:
                    for jj in neighbor_inds_mat[j]:
                        tmp_sum += dist_mat[ii, jj]
                ret_mat[i, j] = tmp_sum / n_neighbor / n_neighbor
                ret_mat[j, i] = ret_mat[i, j]
        return ret_mat

    n_point, n_neighbor = 10, 4
    dist_mat = np.random.randint(low=0, high=100, size=(n_point, n_point))
    dist_mat = np.triu(dist_mat, 1)
    dist_mat += np.transpose(dist_mat)

    new_mat1 = knn_smooth1(dist_mat, n_neighbor)
    new_mat = knn_smooth(dist_mat, n_neighbor)
    print(f"{new_mat= }")
    print(f"{new_mat1= }")
    assert np.all((new_mat - new_mat1) < 1e-4)

def test_merge_consensus_seqs():
    # Test 1
    kmers_example = ["ACGTACGT", "CGTACGT", "TACGTT", "ACGT", "TAC", "CGTA", "ACG","CCTAGGGG", "CTAGGGG", "TAGGGG", "AGG", "GG"]
    res = merge_consensus_seqs(kmers_example)
    assert res == ["CGTACGT", "CTAGGGG"]
    # Test 2


def test_kmap():
    sample_kmer_kh, sample_kmer_count, sample_kmer_label, consensus_list = test_sample_disp_kmer_gen_data()
    kmer_len = 12
    hamdist_mat = cal_samp_kmer_hamdist_mat(sample_kmer_kh, sample_kmer_count, sample_kmer_label, consensus_list, kmer_len, uniq_dist_flag=False)
    ld_data = kmap(hamdist_mat, kmer_len, n_neighbour=20, n_max_iter=100)
    ld_label_arr = _convert_to_block_arr(sample_kmer_label, sample_kmer_count)
    plot_2d_data(ld_data, ld_label_arr, consensus_list)


def test_ex_hamball_kh_arr():
    res_dir = "../test"
    conseq = "CGGAGAGAGTA"
    motif_def_file = "../src/kmap/default_motif_def_table.csv"
    kh_arr, kh_cnt = ex_hamball_kh_arr(res_dir, conseq, motif_def_file=motif_def_file, revcom_mode = True)
    for kh, cnt in zip(kh_arr, kh_cnt):
        print(f"{hash2kmer(kh, len(conseq))}  cnt={cnt}")


def test_ex_hamball():
    res_dir = "./test"
    conseq = "CCATCCATCCATCCA"
    for return_type in ("hash", "kmer", "matrix"):
        output_file = res_dir + "/" + "hamball_" + return_type + ".csv"
        ex_hamball(res_dir, conseq, return_type, output_file)
    draw_logo(res_dir + "/" + "hamball_" + "matrix" + ".csv", res_dir + "/" + "logo.pdf")


def test_scan_motif():
    input_fasta_file = "./tests/test.fa"
    res_dir = "./test"
    debug = False

    _preproc(input_fasta_file, res_dir, debug=debug)
    _scan_motif(res_dir, debug=debug)


def test_visualize_kmers():
    res_dir = "./test"
    debug = True
    _visualize_kmers(res_dir, debug=debug)


if __name__ == "__main__":
    #test_proc_input()
    #test_read_default_config_file()
    #test_comp_kmer_hash_taichi()
    #test_count_uniq_hash()
    #test_mask_input()
    #test_merge_revcom()
    #test_hash2kmer()
    #test_kmer2hash()

    #test_find_motif()
    #test_init_motif_def_dict()
    #test_mask_ham_ball()

    #test_main()
    #test_find_motif()
    #test_sample_disp_kmer()

    #test_sample_disp_kmer_gen_data()

    #test_convert_to_block_mat()
    #test_cal_samp_kmer_hamdist_mat()
    #test_gen_motif_co_occurence_mat()
    #test_knn_smooth()

    #test_merge_consensus_seqs()
    #test_kmap()

    #gen_test_fa_file()
    test_preproc()
    #test_ex_hamball_kh_arr()
    #test_ex_hamball()
    test_scan_motif()
    #test_visualize_kmers()



import taichi as ti

@ti.func
def kmer2hash_taichi_uint32(arr: ti.types.ndarray(dtype=ti.u8), arr_size: int, st_pos: int, k: int,
                            hash_arr: ti.types.ndarray(dtype=ti.u32),
                            invalid_hash: ti.types.u32,
                            missing_val: ti.types.u32):
    # hash_arr[st_pos] = invalid_hash
    invalid_hash_flag = 0
    if st_pos + k > arr_size:
        invalid_hash_flag = 1

    kh = ti.u32(0)
    for i in range(k):
        if arr[st_pos + i] == missing_val:
            invalid_hash_flag = 1
        kh = kh << 2
        kh += arr[st_pos + i]
    hash_arr[st_pos] = kh

    if invalid_hash_flag > 0:
        hash_arr[st_pos] = invalid_hash


@ti.kernel
def kmer2hash_kernel_uint32(arr: ti.types.ndarray(dtype=ti.u8), arr_size: int, k: int,
                            hash_arr: ti.types.ndarray(dtype=ti.u32),
                            invalid_hash_arr: ti.types.ndarray(dtype=ti.u32),
                            missing_val_arr: ti.types.ndarray(dtype=ti.u8)):
    for i in range(arr_size):
        kmer2hash_taichi_uint32(arr, arr_size, i, k, hash_arr, invalid_hash_arr[0], missing_val_arr[0])


@ti.func
def kmer2hash_taichi_uint64(arr: ti.types.ndarray(dtype=ti.u8), arr_size: int, st_pos: int, k: int,
                            hash_arr: ti.types.ndarray(dtype=ti.u64), invalid_hash: ti.u64,
                            missing_val: ti.u8):
    # hash_arr[st_pos] = invalid_hash
    invalid_hash_flag = 0
    if st_pos + k > arr_size:
        invalid_hash_flag = 1

    kh = ti.u64(0)
    for i in range(k):
        if arr[st_pos + i] == missing_val:
            invalid_hash_flag = 1
        kh = kh << 2
        kh += arr[st_pos + i]
    hash_arr[st_pos] = kh

    if invalid_hash_flag > 0:
        hash_arr[st_pos] = invalid_hash


@ti.kernel
def kmer2hash_kernel_uint64(arr: ti.types.ndarray(dtype=ti.u8), arr_size: int, k: int,
                            hash_arr: ti.types.ndarray(dtype=ti.u64),
                            invalid_hash_arr: ti.types.ndarray(dtype=ti.u64),
                            missing_val_arr: ti.types.ndarray(dtype=ti.u8)):
    for i in range(arr_size):
        kmer2hash_taichi_uint64(arr, arr_size, i, k, hash_arr, invalid_hash_arr[0], missing_val_arr[0])

@ti.func
def cal_ham_dist_uint32(hash1: ti.u32, hash2: ti.u32, kmer_len: int):
    xor_result = hash1 ^ hash2
    twobit_mask = ti.cast(3, ti.u32)
    hamming_dist = 0
    for _ in range(kmer_len):
        cmp_res = xor_result & twobit_mask
        hamming_dist += cmp_res != 0
        xor_result >>= 2
    return hamming_dist


@ti.kernel
def cal_ham_dist_kernel_uint32(hash_arr: ti.types.ndarray(dtype=ti.u32),
                               target_hash: ti.types.ndarray(dtype=ti.u32),
                               ham_dist_arr: ti.types.ndarray(dtype=ti.u8),
                               hash_arr_size: int,
                               kmer_len: int):
    for i in range(hash_arr_size):
        ham_dist_arr[i] = ti.cast(cal_ham_dist_uint32(hash_arr[i], target_hash[0], kmer_len), ti.u8)


@ti.func
def cal_ham_dist_uint64(hash1: ti.u64, hash2: ti.u64, kmer_len: int):
    xor_result = hash1 ^ hash2
    twobit_mask = ti.cast(3, ti.u64)
    hamming_dist = 0
    for _ in range(kmer_len):
        cmp_res = xor_result & twobit_mask
        hamming_dist += cmp_res != 0
        xor_result >>= 2
    return hamming_dist


@ti.kernel
def cal_ham_dist_kernel_uint64(hash_arr: ti.types.ndarray(dtype=ti.u64),
                               target_hash: ti.types.ndarray(dtype=ti.u64),
                               ham_dist_arr: ti.types.ndarray(dtype=ti.u8),
                               hash_arr_size: int,
                               kmer_len: int):
    for i in range(hash_arr_size):
        ham_dist_arr[i] = ti.cast(cal_ham_dist_uint64(hash_arr[i], target_hash[0], kmer_len), ti.u8)


# newly added on 11.10.2023, not tested yet, -------------- begin ----------------------------
@ti.func
def cal_partial_ham_dist_head_uint32(kmer_hash: ti.u32, conseq_kh: ti.u32, kmer_len: int, conseq_len: int):
    # match kmer_hash with conseq_kh from the head, kmer_len is greater than conseq_len
    for _ in range(kmer_len - conseq_len):
        kmer_hash >>= 2
    return cal_ham_dist_uint32(kmer_hash, conseq_kh, conseq_len)


@ti.kernel
def cal_partial_ham_dist_head_kernel_uint32(hash_arr: ti.types.ndarray(dtype=ti.u32),
                               target_hash: ti.types.ndarray(dtype=ti.u32),
                               ham_dist_arr: ti.types.ndarray(dtype=ti.u8),
                               hash_arr_size: int,
                               kmer_len: int,
                               conseq_len: int):
    for i in range(hash_arr_size):
        ham_dist_arr[i] = ti.cast(cal_partial_ham_dist_head_uint32(hash_arr[i], target_hash[0], kmer_len, conseq_len), ti.u8)


@ti.func
def cal_partial_ham_dist_tail_uint32(kmer_hash: ti.u32, conseq_kh: ti.u32, kmer_len: int, conseq_len: int):
    # match kmer_hash with conseq_kh from the head, kmer_len is greater than conseq_len
    return cal_ham_dist_uint32(kmer_hash, conseq_kh, conseq_len)


@ti.kernel
def cal_partial_ham_dist_tail_kernel_uint32(hash_arr: ti.types.ndarray(dtype=ti.u32),
                               target_hash: ti.types.ndarray(dtype=ti.u32),
                               ham_dist_arr: ti.types.ndarray(dtype=ti.u8),
                               hash_arr_size: int,
                               kmer_len: int,
                               conseq_len: int):
    for i in range(hash_arr_size):
        ham_dist_arr[i] = ti.cast(cal_partial_ham_dist_tail_uint32(hash_arr[i], target_hash[0], kmer_len, conseq_len), ti.u8)


@ti.func
def cal_partial_ham_dist_head_uint64(kmer_hash: ti.u64, conseq_kh: ti.u64, kmer_len: int, conseq_len: int):
    # match kmer_hash with conseq_kh from the head, kmer_len is greater than conseq_len
    for _ in range(kmer_len - conseq_len):
        kmer_hash >>= 2
    return cal_ham_dist_uint64(kmer_hash, conseq_kh, conseq_len)


@ti.kernel
def cal_partial_ham_dist_head_kernel_uint64(hash_arr: ti.types.ndarray(dtype=ti.u64),
                               target_hash: ti.types.ndarray(dtype=ti.u64),
                               ham_dist_arr: ti.types.ndarray(dtype=ti.u8),
                               hash_arr_size: int,
                               kmer_len: int,
                               conseq_len: int):
    for i in range(hash_arr_size):
        ham_dist_arr[i] = ti.cast(cal_partial_ham_dist_head_uint64(hash_arr[i], target_hash[0], kmer_len, conseq_len), ti.u8)


@ti.func
def cal_partial_ham_dist_tail_uint64(kmer_hash: ti.u64, conseq_kh: ti.u64, kmer_len: int, conseq_len: int):
    # match kmer_hash with conseq_kh from the head, kmer_len is greater than conseq_len
    return cal_ham_dist_uint64(kmer_hash, conseq_kh, conseq_len)


@ti.kernel
def cal_partial_ham_dist_tail_kernel_uint64(hash_arr: ti.types.ndarray(dtype=ti.u64),
                               target_hash: ti.types.ndarray(dtype=ti.u64),
                               ham_dist_arr: ti.types.ndarray(dtype=ti.u8),
                               hash_arr_size: int,
                               kmer_len: int,
                               conseq_len: int):
    for i in range(hash_arr_size):
        ham_dist_arr[i] = ti.cast(cal_partial_ham_dist_tail_uint64(hash_arr[i], target_hash[0], kmer_len, conseq_len), ti.u8)

# newly added on 11.10.2023, not tested yet, -------------- end ----------------------------

@ti.func
def revcom_hash_uint32(in_hash: ti.u32,
                       mask: ti.u32,
                       twobit_mask: ti.u32,
                       k: int):
    com_hash = mask - in_hash  # complement hash
    ret_hash = twobit_mask & com_hash
    for i in range(k - 1):
        ret_hash = ret_hash << 2
        com_hash = com_hash >> 2
        ret_hash += twobit_mask & com_hash
    return ret_hash


@ti.func
def revcom_hash_uint64(in_hash: ti.u64,
                       mask: ti.u64,
                       twobit_mask: ti.u64,
                       k: int):
    com_hash = mask - in_hash  # complement hash
    ret_hash = twobit_mask & com_hash
    for i in range(k - 1):
        ret_hash = ret_hash << 2
        com_hash = com_hash >> 2
        ret_hash += twobit_mask & com_hash
    return ret_hash


@ti.kernel
def revcom_hash_kernel_uint32(in_hash_arr: ti.types.ndarray(dtype=ti.u32),
                              out_hash_arr: ti.types.ndarray(dtype=ti.u32),
                              mask_arr: ti.types.ndarray(dtype=ti.u32),
                              kmer_len: int, in_hash_arr_size: int):
    for i in range(in_hash_arr_size):
        out_hash_arr[i] = revcom_hash_uint32(in_hash_arr[i], mask_arr[0], mask_arr[1], kmer_len)


@ti.kernel
def revcom_hash_kernel_uint64(in_hash_arr: ti.types.ndarray(dtype=ti.u64),
                              out_hash_arr: ti.types.ndarray(dtype=ti.u64),
                              mask_arr: ti.types.ndarray(dtype=ti.u64),
                              kmer_len: int, in_hash_arr_size: int):
    for i in range(in_hash_arr_size):
        out_hash_arr[i] = revcom_hash_uint64(in_hash_arr[i], mask_arr[0], mask_arr[1], kmer_len)


@ti.func
def cal_avg_dist(dist_mat: ti.types.ndarray(dtype=ti.f32),
                 neighbor_inds_mat: ti.types.ndarray(dtype=ti.i32),
                 new_dist_mat: ti.types.ndarray(dtype=ti.f32),
                 i: int, j: int, n_neighbour: int):
    tmp_sum = 0.0
    for ii in range(n_neighbour):
        for jj in range(n_neighbour):
            tmp_sum += dist_mat[neighbor_inds_mat[i, ii], neighbor_inds_mat[j, jj]]
    new_dist_mat[i, j] = tmp_sum / n_neighbour / n_neighbour


@ti.kernel
def knn_smooth_kernel(dist_mat: ti.types.ndarray(dtype=ti.f32),
                      neighbor_inds_mat: ti.types.ndarray(dtype=ti.i32),
                      new_dist_mat: ti.types.ndarray(dtype=ti.f32),
                      n_point: int, n_neighbour: int):
    for k in range(n_point * n_point):
        i = k // n_point
        j = k % n_point
        if i >= j:
            continue
        cal_avg_dist(dist_mat, neighbor_inds_mat, new_dist_mat, i, j, n_neighbour)


@ti.func
def cal_ld_prob_pair(xi: ti.f32, xj: ti.f32, yi: ti.f32, yj: ti.f32):
    dist2 = (xi - xj) * (xi - xj) + (yi - yj) * (yi - yj)
    dist2 = 1 / (1 + dist2)
    return dist2


@ti.kernel
def cal_ld_prob_mat_kernel(x_arr: ti.types.ndarray(dtype=ti.f32),
                           y_arr: ti.types.ndarray(dtype=ti.f32),
                           n_iter: int,
                           i_ind_arr: ti.types.ndarray(dtype=ti.i32),
                           j_ind_arr: ti.types.ndarray(dtype=ti.i32),
                           prob_mat: ti.types.ndarray(dtype=ti.f32)):
    for ind in range(n_iter):
        i = i_ind_arr[ind]
        j = j_ind_arr[ind]
        prob_mat[i, j] = cal_ld_prob_pair(x_arr[i], x_arr[j], y_arr[i], y_arr[j])
        prob_mat[j, i] = prob_mat[i, j]

@ti.kernel
def cal_cross_entropy_kernel(hd_prob_mat: ti.types.ndarray(dtype=ti.f32),
                             ld_prob_mat: ti.types.ndarray(dtype=ti.f32),
                             n_iter: int,
                             i_ind_arr: ti.types.ndarray(dtype=ti.i32),
                             j_ind_arr: ti.types.ndarray(dtype=ti.i32),
                             cross_entropy_mat: ti.types.ndarray(dtype=ti.f32)):
    eps = 1e-10
    ld_prob = 0.0
    for ind in range(n_iter):
        i, j = i_ind_arr[ind], j_ind_arr[ind]
        if ld_prob_mat[i, j] < eps:
            ld_prob = eps
        elif ld_prob_mat[i, j] > 1 - eps:
            ld_prob = 1 - eps
        else:
            ld_prob = ld_prob_mat[i, j]

        ret_val = 0.0
        if hd_prob_mat[i, j] < eps:
            ret_val = - ti.log(1-ld_prob)
        elif hd_prob_mat[i, j] > 1 - eps:
            ret_val = - ti.log(ld_prob)
        else:
            ret_val = cal_cross_entropy(hd_prob_mat[i, j], ld_prob)

        cross_entropy_mat[i, j] = ret_val


@ti.func
def cal_cross_entropy(hd_prob: ti.f32, ld_prob: ti.f32):
    return - hd_prob * ti.log(ld_prob) - (1 - hd_prob) * ti.log(1 - ld_prob)

@ti.kernel
def cal_gradient_loss_kernel(tmp_diff_mat: ti.types.ndarray(dtype=ti.f32), # n x n
                             ld_data: ti.types.ndarray(dtype=ti.f32), # 2 x n
                             ret_mat: ti.types.ndarray(dtype=ti.f32), # 2 x n
                             n: int):
    for i in range(n):
        ret_mat[0, i] = cal_gradient_loss(tmp_diff_mat, ld_data, 0, n, i)
        ret_mat[1, i] = cal_gradient_loss(tmp_diff_mat, ld_data, 1, n, i)

@ti.func
def cal_gradient_loss(diff_mat: ti.types.ndarray(dtype=ti.f32),
                      ld_data: ti.types.ndarray(dtype=ti.f32),
                      k: int, # kth dimension
                      n: int,
                      i: int):
    ret_val = 0.0
    for j in range(n):
        if i == j:
            continue
        else:
            ret_val += diff_mat[i, j] * (ld_data[k, i] - ld_data[k, j])
    return ret_val




import matplotlib.pyplot as plt

from .kmer_count import (comp_kmer_hash_taichi, count_uniq_hash, merge_revcom,
                                 cal_hamming_dist, revcom_hash, mask_input, proc_input,
                                 init_motif_def_dict, mask_ham_ball, hash2kmer, kmer2hash,
                                 cal_hamming_dist_head, cal_hamming_dist_tail, dna2arr,
                                 reverse_complement, get_hash_dtype, FileNameDict, gen_motif_def_dict)
import numpy as np
import pickle
from scipy.stats import norm
from typing import List, Tuple
import warnings

from pathlib import Path
from Bio import SeqIO
import gzip
import tomllib
import logomaker
import pandas as pd
import click


@click.command(name="scan_motif")
@click.option(
    '--res_dir',
    type=str,
    help='Result directory for storing all outputs',
    required=True
    )
@click.option(
    '--debug',
    type=bool,
    default=False,
    help='display debug information.',
    required=False
    )
def scan_motif(res_dir: str, debug=False):
    _scan_motif(res_dir, debug)


@click.command(name="draw_logo")
@click.option(
    '--cnt_mat_numpy_file',
    type=str,
    help='count matrix file',
    required=True
    )
@click.option(
    '--output_fig_file',
    type=str,
    default="motif_logo.png",
    help='output figure file name.',
    required=False
    )
def draw_logo(cnt_mat_numpy_file: str, output_fig_file=None):
    _draw_logo(cnt_mat_numpy_file, output_fig_file)


@click.command(name="ex_hamball")
@click.option(
    '--res_dir',
    type=str,
    help='Result directory for storing all outputs',
    required=True
    )
@click.option(
    '--conseq',
    type=str,
    help='the consensus sequence',
    required=True
    )
@click.option(
    '--return_type',
    type=str,
    help='output file form, can be ["hash" | "kmer" | "matrix"]',
    required=True
    )
@click.option(
    '--output_file',
    type=str,
    help='output file name, including the suffix',
    required=True
    )
@click.option(
    '--max_ham_dist',
    type=int,
    default=-1,
    help='The radius of the Hamming ball. -1 means taking the radius from motif_def_table.csv',
    required=False
    )
def ex_hamball(res_dir: str, conseq: str, return_type: str, output_file: str,
               max_ham_dist: int=-1):
    _ex_hamball(res_dir, conseq, return_type, output_file, max_ham_dist)


def write_lines(str_list: List, outfile: str|Path):
    with open(outfile, 'w+') as fh:
        for line in str_list:
            fh.write(line + "\n")


def _scan_motif(res_dir: str, debug=False):
    config_file_name = FileNameDict["config_file"] # config.toml
    config_file_path = Path(res_dir) / config_file_name

    motif_def_file = FileNameDict["motif_def_file"] # motif_def_table.csv
    motif_def_file_path = Path(res_dir) / motif_def_file

    proc_fasta_file = FileNameDict["processed_fasta_file"] # input.bin.pkl
    proc_fasta_file_path = Path(res_dir) / proc_fasta_file

    assert config_file_path.exists()
    assert motif_def_file_path.exists()
    assert proc_fasta_file_path.exists()

    # load config and motif_def files
    with open(config_file_path, "rb") as fh:
        config_dict = tomllib.load(fh)
    motif_def_dict = gen_motif_def_dict(config_dict, debug=debug)
    min_k = config_dict["kmer_count"]["min_k"]
    max_k = config_dict["kmer_count"]["max_k"]
    revcom_mode = config_dict["kmer_count"]["revcom_mode"]

    mask_noise_seq_list = []
    if config_dict["motif_discovery"]["noise_kmer_file"] != "None":
        noise_kmer_file = config_dict["motif_discovery"]["noise_kmer_file"]
        assert Path(noise_kmer_file).exists()
        with open(Path(noise_kmer_file), "r") as fh:
            for line in fh:
                line = line.strip()
                if len(line) > 0:
                    mask_noise_seq_list.append(line)

    with open(proc_fasta_file_path, "rb") as fh:
        seq_np_arr = pickle.load(fh)

    # mask user provided noise sequences from the input sequence
    if len(mask_noise_seq_list) > 0:
        max_ham_dist_list = [0 for _ in mask_noise_seq_list]
        seq_np_arr = mask_ham_ball(seq_np_arr, motif_def_dict, mask_noise_seq_list, max_ham_dist_list)

    # load necessary parameters
    top_k = config_dict["motif_discovery"]["top_k"]
    n_trial = config_dict["motif_discovery"]["n_trial"]
    save_kmer_cnt_flag = config_dict["motif_discovery"]["save_kmer_cnt_flag"]
    orig_seq_np_arr = seq_np_arr.copy()
    candidate_conseq_list = []

    if save_kmer_cnt_flag:
        kmer_count_dir = Path(res_dir) / FileNameDict["kmer_count_dir"]
        if not kmer_count_dir.exists():
            kmer_count_dir.mkdir()

    # candidate motif sequences
    candidate_conseq_file = Path(res_dir) / FileNameDict["candidate_conseq_file"]
    if candidate_conseq_file.exists():
        print(f"{candidate_conseq_file} already exist, re-use it.")
    else:
        # motif discovery
        res = ["kmer_len,conseq_hash,conseq,hamball_proportion,hamball_ratio,log10_p_value"]
        for kmer_len in range(min_k, max_k + 1):
            seq_np_arr[:] = orig_seq_np_arr[:]
            p_uniform_k = motif_def_dict[kmer_len].p_uniform
            max_ham_dist = motif_def_dict[kmer_len].max_ham_dist
            ratio_mu = motif_def_dict[kmer_len].ratio_mu
            ratio_std = motif_def_dict[kmer_len].ratio_std
            ratio_cutoff = motif_def_dict[kmer_len].ratio_cutoff

            kmer_cnt_file = Path(res_dir) / FileNameDict[
                "kmer_count_dir"] / f"k{kmer_len}.pkl"  # [kmer_len, uniq_kh_arr, uniq_kh_cnt_arr]
            consensus_kh_dict = find_motif(seq_np_arr, kmer_len, max_ham_dist, p_uniform_k,
                                           ratio_mu, ratio_std, ratio_cutoff,
                                           top_k, n_trial,
                                           revcom_mode, save_kmer_cnt_flag=save_kmer_cnt_flag,
                                           kmer_cnt_pkl_file=kmer_cnt_file,
                                           debug=debug)  # snp_np_arr are mutated in find_motif()
            if debug:
                print(f"filtered consensus kmers when k = {kmer_len}")
            for kh, values in consensus_kh_dict.items():
                prop, ratio, log10_p_value = values
                kmer_seq = hash2kmer(kh, kmer_len)
                if debug:
                    print(
                        f'{hash2kmer(kh, kmer_len)} perc={prop * 100:0.5f}%  hamball_ratio={ratio} log10_p_value={log10_p_value}')
                res.append(f"{kmer_len},{kh},{kmer_seq},{prop:0.8f},{ratio:0.4f},{log10_p_value:0.4f}")
                candidate_conseq_list.append(kmer_seq)
        print(f"kmer counting finished for k={min_k}...{max_k}. Candidate consensus sequences generated.")
        write_lines(res, candidate_conseq_file)

    # merge candidate motif sequences
    final_conseq_file = Path(res_dir) / FileNameDict["final_conseq_file"]
    if final_conseq_file.exists():
        with open(final_conseq_file, "r") as fh:
            final_conseq_list = fh.read().splitlines()
        print(f"{final_conseq_file} already exist, re-use it.")
    else:
        final_conseq_list = merge_consensus_seqs(candidate_conseq_list)
        write_lines(final_conseq_list, final_conseq_file)

    final_conseq_info_file = Path(res_dir) / FileNameDict["final_conseq_info_file"]
    if final_conseq_info_file.exists():
        print(f"{final_conseq_info_file} already exist, re-use it.")
    else:
        with open(final_conseq_file, "r") as fh:
            final_conseq_list = fh.read().splitlines()
        with open(candidate_conseq_file, "r") as fh:
            candidate_conseq_info_list = fh.read().splitlines()
        final_conseq_info_list = [candidate_conseq_info_list[0]] # header
        for conseq in final_conseq_list:
            for line in candidate_conseq_info_list:
                if ","+conseq+"," in line:
                    final_conseq_info_list.append(line)
                    continue
        write_lines(final_conseq_info_list, final_conseq_info_file)
    print("Final consensus sequences generated.")

    # generate motif co-occurence matrix
    co_occur_mat_file = Path(res_dir) / FileNameDict["co_occur_mat_file"]
    co_occur_mat_norm_file = Path(res_dir) / FileNameDict["co_occur_mat_norm_file"]
    if co_occur_mat_file.exists():
        print(f"{co_occur_mat_file}, re-use it!")
    else:
        min_motif_cnt = config_dict["motif_discovery"]["min_motif_cnt"]
        input_fasta_file = config_dict["general"]["input_fasta_file"]
        co_occur_mat = gen_motif_co_occurence_mat(input_fasta_file, final_conseq_list,
                                                  motif_def_dict, min_motif_cnt=min_motif_cnt, revcom_mode=revcom_mode)
        co_sum_mat = np.diag(co_occur_mat) + np.diag(co_occur_mat).reshape((-1, 1))
        co_occur_norm_mat = co_occur_mat / (co_sum_mat - co_occur_mat)
        with open(co_occur_mat_file, "w+") as fh:
            np.savetxt(fh, co_occur_mat, delimiter="\t", fmt="%d")
        with open(co_occur_mat_norm_file, "w+") as fh:
            np.savetxt(fh, co_occur_norm_mat, delimiter="\t", fmt="%2.3f")
    print("motif co-occurence matrix generated.")

    # sample kmers
    if not save_kmer_cnt_flag:
        print(f"kmers cannot be sampled when {save_kmer_cnt_flag=}, quit!")
        return
    n_total_sample = config_dict["motif_discovery"]["n_total_sample"]
    n_motif_sample = config_dict["motif_discovery"]["n_motif_sample"]
    kmer_count_dir = Path(res_dir) / FileNameDict["kmer_count_dir"]
    kmer_len = max([len(conseq) for conseq in final_conseq_list])
    (samp_kh_arr, samp_cnts, samp_label_arr, conseq_list) \
        = sample_disp_kmer(final_conseq_list, kmer_len, motif_def_dict,
            kmer_count_dir=kmer_count_dir, n_total_sample = n_total_sample,
            n_motif_kmer = n_motif_sample, revcom_mode = revcom_mode)
    sample_kmer_pkl_file = Path(res_dir) / FileNameDict["sample_kmer_pkl_file"]
    sample_kmer_txt_file = Path(res_dir) / FileNameDict["sample_kmer_txt_file"]
    with open(sample_kmer_pkl_file, "wb") as fh:
        pickle.dump([samp_kh_arr, samp_cnts, samp_label_arr, conseq_list], fh)
    sample_kmer_lines = []
    for kh, cnt, label in zip(samp_kh_arr, samp_cnts, samp_label_arr):
        for _ in range(cnt):
            sample_kmer_lines.append(f"{hash2kmer(kh, kmer_len)}\t{label}")
    write_lines(sample_kmer_lines, sample_kmer_txt_file)
    print(f"kmers are sampled for visualization. {kmer_len= }, {n_total_sample= }, {n_motif_sample= }")

    # calculate Hamming distance between sampled kmers
    hamdist_mat = cal_samp_kmer_hamdist_mat(samp_kh_arr, samp_cnts, samp_label_arr, conseq_list, kmer_len,
                                            uniq_dist_flag = False)
    label_arr = _convert_to_block_arr(samp_label_arr, samp_cnts)
    sample_kmer_hamdist_mat_file = Path(res_dir) / FileNameDict["sample_kmer_hamdist_mat_file"]
    with open(sample_kmer_hamdist_mat_file, "wb") as fh:
        pickle.dump([kmer_len, hamdist_mat, label_arr], fh)
    print(f"Hamming distance matrix of sampled kmers are generated.")

    gen_hamball_flag = config_dict["motif_discovery"]["gen_hamball_flag"]
    if gen_hamball_flag:
        for i, conseq in enumerate(final_conseq_list):
            if debug:
                print(f"generating motif count matrix and draw logo for motif {i}: {conseq}")
            out_dir_path = Path(res_dir) / FileNameDict["hamball_dir"]
            if not out_dir_path.exists():
                out_dir_path.mkdir()
            return_type = "matrix"
            output_cntmat_file = str(out_dir_path / f"cntmat_motif{i}_{conseq}.csv")
            max_ham_dist = motif_def_dict[len(conseq)].max_ham_dist
            _ex_hamball(res_dir, conseq, return_type, output_cntmat_file, max_ham_dist=max_ham_dist)
            output_logo_file = str(out_dir_path / f"logo_motif{i}_{conseq}.pdf")
            _draw_logo(output_cntmat_file, output_fig_file = output_logo_file)
        print("Motif count matrix and logo extracted.")

    print(f"All tasks of scan motif finished.")


def _ex_hamball(res_dir: str, conseq: str, return_type: str, output_file: str,
               max_ham_dist: int=-1):
    """
    Extract kmers of a Hamming ball
    Args:
        res_dir: result directory
        conseq: consensus sequence
        return_type: "hash" | "kmer" | "matrix"
        output_file: output file name
        max_ham_dist: maximum Hamming distance of the Hamming ball
        revcom_mode: reverse complement mode
    Returns:
        Tuple (hash,cnt) | (kmer, cnt) | matrix
    """
    config_file_name = FileNameDict["config_file"]  # config.toml
    config_file_path = Path(res_dir) / config_file_name
    assert config_file_path.exists()

    # load config and motif_def files
    with open(config_file_path, "rb") as fh:
        config_dict = tomllib.load(fh)

    assert return_type in ("hash", "kmer", "matrix")
    motif_def_file = FileNameDict["motif_def_file"]  # motif_def_table.csv
    motif_def_file_path = Path(res_dir) / motif_def_file
    revcom_mode = config_dict["kmer_count"]["revcom_mode"]

    uniq_kh_arr, uniq_kh_cnt_arr = ex_hamball_kh_arr(res_dir, conseq, max_ham_dist, motif_def_file_path, revcom_mode)
    kmer_len = len(conseq)

    with open(output_file, 'w+') as fh:
        if return_type == "hash":
            for kh, cnt in zip(uniq_kh_arr, uniq_kh_cnt_arr):
                fh.write(f"{kh},{cnt}\n")
        elif return_type == "kmer":
            for kh, cnt in zip(uniq_kh_arr, uniq_kh_cnt_arr):
                fh.write(f"{hash2kmer(kh, kmer_len)},{cnt}\n")
        else:
            cnt_mat = cal_cnt_mat(uniq_kh_arr, uniq_kh_cnt_arr, kmer_len)
            np.savetxt(fh, cnt_mat, delimiter=",", fmt="%d")

    print(f"Extract Hamming ball [type={return_type}] save in {output_file}.")


def merge_consensus_seqs(conseq_list: List[str]) -> List[str]:
    """
    merge motifs of different lengths
    Args:
        conseq_list: all candidate motif consensus sequences of different lengths
    Returns:
        final consensus sequences
    """

    def _overlap(long_kmer, short_kmer):
        # if short_kmer is a substring of long_kmer
        len_l, len_s = len(long_kmer), len(short_kmer)
        assert len_l >= len_s
        for i in range(len_l - len_s + 1):
            if short_kmer == long_kmer[i:i+len_s]:
                return True
        return False

    def _overlap_shift_one(long_kmer, short_kmer):
        # if k-1 substring of short_kmer is also a subtring of long_kmer
        return _overlap(long_kmer, short_kmer[:-1]) or _overlap(long_kmer, short_kmer[1:])

    # Sort the k-mers in descending order of their length
    conseq_list = sorted(conseq_list, key=len, reverse=True)
    final_conseq_list = []

    while len(conseq_list) > 0:
        curr_conseq = conseq_list[0]
        rc_curr_conseq = reverse_complement(curr_conseq)

        conseq_len_list = [len(conseq) for conseq in conseq_list]
        sub_inds_1 = [i for i, seq_len in enumerate(conseq_len_list) if seq_len == (len(curr_conseq) - 1)]
        sub_inds_2 = [i for i, seq_len in enumerate(conseq_len_list) if seq_len == (len(curr_conseq) - 2)]

        substr1 = None
        for i1 in sub_inds_1:
            if _overlap_shift_one(curr_conseq, conseq_list[i1]) or _overlap_shift_one(rc_curr_conseq, conseq_list[i1]):
                substr1 = conseq_list[i1]
                break
        substr2 = None
        for i2 in sub_inds_2:
            if _overlap_shift_one(curr_conseq, conseq_list[i2]) or _overlap_shift_one(rc_curr_conseq, conseq_list[i2]):
                substr2 = conseq_list[i2]
                break

        if substr1 and substr2:
            final_conseq_list.append(substr1)
            # remove all substrings
            new_conseq_list = []
            for i, conseq in enumerate(conseq_list):
                if _overlap_shift_one(curr_conseq, conseq) or _overlap_shift_one(rc_curr_conseq, conseq):
                    continue
                else:
                    new_conseq_list.append(conseq)
            conseq_list = new_conseq_list
        else:
            conseq_list = conseq_list[1:]

    return final_conseq_list


def find_motif(seq_np_arr, kmer_len: int, max_ham_dist, p_unif,
               ratio_mu, ratio_std, ratio_cutoff, top_k=5, n_trial=10,
               merge_revcom_mode=True, save_kmer_cnt_flag=True,
               kmer_cnt_pkl_file: Path=None, debug=False) -> dict:
    """
    main motif discovery code,
        step 0: we try to pick the largest hamming ball for top k kmers each time,
        step 1: check if it passes the significant test
        step 3: repeat the process n_trial times
    Args:
        seq_np_arr: input sequence numpy array (uint8), missing values are 255
        kmer_len: kmer length
        max_ham_dist: maximum hamming ball distance for the given kmer length
        p_uniform: probability of a hamming ball centered on a random kmer
        ratio_mu: mean of the Hamming ball ratio distribution
        ratio_std: std of the Hamming ball ratio distribution
        ratio_cutoff: Hamming ball ratio cutoff for a kmer to be considered as significant
        top_k: top k consensus sequences to consider
        n_trial: number of times we try to pick up a motif
        merge_revcom_mode: if revcom should be merged
    Returns:
        dict, key is consensus_kmer_hash, value is a tuple of the Hamming ball (proportion, ratio, log10_pvalue)
    """

    if save_kmer_cnt_flag and kmer_cnt_pkl_file and Path(kmer_cnt_pkl_file).exists():
        with open(Path(kmer_cnt_pkl_file), "rb") as fh:
            kmer_len_from_pkl_file, uniq_kh_arr, uniq_kh_cnt_arr = pickle.load(fh)
            assert kmer_len == kmer_len_from_pkl_file
    else:
        # first round
        hash_arr = comp_kmer_hash_taichi(seq_np_arr, kmer_len)
        uniq_kh_arr, uniq_kh_cnt_arr = count_uniq_hash(hash_arr, kmer_len)
        # merge revcom
        if merge_revcom_mode:
            uniq_kh_arr, uniq_kh_cnt_arr = merge_revcom(uniq_kh_arr, uniq_kh_cnt_arr, kmer_len,
                                                        keep_lower_hash_flag=True)

    if save_kmer_cnt_flag and kmer_cnt_pkl_file and not Path(kmer_cnt_pkl_file).exists():
        assert kmer_cnt_pkl_file
        with open(kmer_cnt_pkl_file, "wb") as fh:
           pickle.dump([kmer_len, uniq_kh_arr, uniq_kh_cnt_arr], fh)

    # count total kmer excluding invalid kmer
    n_total_kmer = sum(uniq_kh_cnt_arr)

    res_consensus_kh_list = []
    res_consensus_proportion_list = []
    res_consensus_ratio_list = []
    res_consensus_log10_pvalue_list =[]

    for i_trial in range(n_trial):
        # get the kmer with maximum hamming ball counts
        top_k_inds = np.array(np.argpartition(uniq_kh_cnt_arr, -top_k)[-top_k:])
        if len(top_k_inds) == 0:
            break

        hamball_cnt_arr = np.zeros(top_k)
        for i, ind in enumerate(top_k_inds):
            kh = uniq_kh_arr[ind]
            dist_arr = cal_hamming_dist(uniq_kh_arr, kh, kmer_len)
            if merge_revcom_mode:
                rc_kh = revcom_hash(kh, kmer_len)
                rc_dist_arr = cal_hamming_dist(uniq_kh_arr, rc_kh, kmer_len)  # revcom
                dist_arr = np.minimum(dist_arr, rc_dist_arr)
            hamball_cnt_arr[i] = np.sum(uniq_kh_cnt_arr[dist_arr <= max_ham_dist])

        if debug:
            print(f"{i_trial= }")

        max_hamball_ind = np.argmax(hamball_cnt_arr)
        consensus_kh = uniq_kh_arr[top_k_inds[max_hamball_ind]]
        hamball_proportion = (hamball_cnt_arr[max_hamball_ind] + 0.0) / n_total_kmer
        hamball_ratio = hamball_proportion / p_unif

        if hamball_ratio > ratio_cutoff:
            res_consensus_kh_list.append(consensus_kh)
            res_consensus_proportion_list.append(hamball_proportion)
            res_consensus_ratio_list.append(hamball_ratio)
            res_consensus_log10_pvalue_list.append( norm.logsf(hamball_ratio, loc=ratio_mu, scale=ratio_std)/np.log(10) )

            if merge_revcom_mode:
                rc_consensus_kh = revcom_hash(consensus_kh, kmer_len)
                seq_np_arr = mask_input(seq_np_arr, kmer_len, np.array([consensus_kh, rc_consensus_kh]), np.array([max_ham_dist, max_ham_dist]))
            else:
                seq_np_arr = mask_input(seq_np_arr, kmer_len, np.array([consensus_kh]), np.array([max_ham_dist]))

            hash_arr = comp_kmer_hash_taichi(seq_np_arr, kmer_len)
            uniq_kh_arr, uniq_kh_cnt_arr = count_uniq_hash(hash_arr, kmer_len)
            if merge_revcom_mode:
                uniq_kh_arr, uniq_kh_cnt_arr = merge_revcom(uniq_kh_arr, uniq_kh_cnt_arr, kmer_len,
                                                            keep_lower_hash_flag=True)
        else:
            break
    return dict(zip(res_consensus_kh_list, zip(res_consensus_proportion_list, res_consensus_ratio_list, res_consensus_log10_pvalue_list)))


def _convert_to_block_mat(uniq_dist_mat: np.ndarray, block_size_arr: np.ndarray) -> np.ndarray:
    """
    convert each element of uniq_dist_mat (square matrix) to a block, with all elements in the block having the same value
    Args:
        uniq_dist_mat: a nq x nq matrix
        block_size_arr: block size for each element in the output
    Returns:
        an expanded matrix
    """

    assert np.issubdtype(block_size_arr.dtype, np.integer)
    assert np.all(block_size_arr > 0)

    # expand the matrix
    boarder_arr = np.zeros(len(block_size_arr) + 1, dtype=int)
    boarder_arr[1:] = np.cumsum(block_size_arr)
    st_arr = boarder_arr[:-1]
    en_arr = boarder_arr[1:]
    n_seq = boarder_arr[-1]
    hamdist_mat = np.zeros((n_seq, n_seq), dtype=uniq_dist_mat.dtype)
    for i in range(len(block_size_arr)):
        for j in range(len(block_size_arr)):
            st_i, en_i = st_arr[i], en_arr[i]
            st_j, en_j = st_arr[j], en_arr[j]
            hamdist_mat[st_i:en_i, st_j:en_j] = uniq_dist_mat[i, j]
    return hamdist_mat


def _convert_to_block_arr(arr: np.ndarray, block_size_arr: np.ndarray) -> np.ndarray:
    """
    convert each element of arr to a block, with all elements in the block having the same value
    Args:
        arr: 1 x n array
        block_size_arr: block size for each element in the output, 1 x n array
    Returns:
        an expanded array
    """

    assert np.issubdtype(block_size_arr.dtype, np.integer)
    assert np.all(block_size_arr > 0)
    assert len(arr) == len(block_size_arr)

    # expand the matrix
    boarder_arr = np.zeros(len(block_size_arr) + 1, dtype=int)
    boarder_arr[1:] = np.cumsum(block_size_arr)
    st_arr = boarder_arr[:-1]
    en_arr = boarder_arr[1:]
    n_seq = boarder_arr[-1]
    out_arr = np.zeros(n_seq, dtype=arr.dtype)
    for i in range(len(block_size_arr)):
        st_i, en_i = st_arr[i], en_arr[i]
        out_arr[st_i:en_i] = arr[i]
    return out_arr

def cal_samp_kmer_hamdist_mat(samp_kh_arr: np.ndarray, samp_cnts: np.ndarray,
                           samp_label_arr: np.ndarray, conseq_list: List[str], kmer_len: int, uniq_dist_flag=False) -> np.ndarray:
    """
    Calculate the hamming distance between the sampled kmer
    When calculating distances for kmers with the same label, we only consider the first n letters, where n=len(conseq)
    Each uniq kmer is expanded uniq_kmer_cnt times
    Args:
        samp_kh_arr: sampled unique kmer hash arr, note that the kmers has been revcom-ed to align with the conseqs
        samp_cnts: counts of sampled kmer hash arr
        samp_label_arr: label of kmer, which conseq it belongs to
        conseq_list: consensus sequence list
        kmer_len: kmer length of the main conseqs
        uniq_dist_flag: if return the distance matrix for samp_kh_arr (unique values),
                        or expand it such that there are samp_cnts[i] replicates for samp_kh_arr[i]
    Returns:
        a hamming distance matrix, with each row representing a kmer
    """

    assert len(samp_kh_arr) == len(np.unique(samp_kh_arr)) # sample_kh_arr must only contain unique values
    n_uniq_kmer = len(samp_kh_arr)
    uniq_dist_mat = np.zeros((n_uniq_kmer, n_uniq_kmer), dtype=int)

    for conseq in conseq_list:
        assert len(conseq) <= kmer_len

    # calculate hamming dist based on kmer_len for all kmers
    for i, kh in enumerate(samp_kh_arr):
        uniq_dist_mat[i, (i+1):] = cal_hamming_dist(samp_kh_arr[(i+1):], kh, kmer_len)
        uniq_dist_mat[(i + 1):, i] = uniq_dist_mat[i, (i+1):]

    # calculate hamming dist for kmers belong to a short conseq
    for i, conseq in enumerate(conseq_list):
        if len(conseq) == kmer_len:
            continue
        tmpinds = np.where(samp_label_arr == i)[0]
        tmp_kh_arr = samp_kh_arr[tmpinds]
        conseq_len = len(conseq)
        tmp_kh_arr = np.right_shift(tmp_kh_arr, 2 * (kmer_len - conseq_len)).astype(get_hash_dtype(conseq_len))
        for j, kh in enumerate(tmp_kh_arr):
            tmp_dist_arr = cal_hamming_dist(tmp_kh_arr[(j + 1):], kh, conseq_len)
            uniq_dist_mat[tmpinds[j], tmpinds[(j + 1):]] = tmp_dist_arr
            uniq_dist_mat[tmpinds[(j + 1):], tmpinds[j]] = tmp_dist_arr

    # expand the matrix
    hamdist_mat = _convert_to_block_mat(uniq_dist_mat, samp_cnts)

    if uniq_dist_flag:
        return uniq_dist_mat
    else:
        return hamdist_mat


# sample display kmers
def sample_disp_kmer(conseq_list: List[str], kmer_len: int, motif_def_dict: dict, kmer_count_dir: Path,
                     n_total_sample=5000, n_motif_kmer=2500, revcom_mode=True) -> Tuple:
    """
    Sample kmers for visualization
    Args:
        conseq_list: consensus sequence list
        kmer_len: kmer length
        motif_def_dict: motif definition tables
        kmer_count_dir: output directory that contains the kmer count result
        n_total_sample: total number of samples for visualization
        n_motif_kmer: number of motif kmers, n_sample-n_motif_kmer is the number of random kmers
        revcom_mode: if reverse complement exist in data

    Returns:
        kmer_hash_arr, kmer_hash_cnt_arr, label_arr, conseq_list
    """
    # process conseq
    conseq_list = [s for s in conseq_list if 2 < len(s) <= kmer_len]
    assert len(conseq_list) > 0
    assert all([len(s_i) >= len(s_i_plus_1) for s_i, s_i_plus_1 in zip(conseq_list, conseq_list[1:])])
    #conseq_list = sorted(conseq_list, key=lambda x: len(x)) # sort conseq by length

    # load kmer counts
    kmer_cnt_file = kmer_count_dir / f"k{kmer_len}.pkl"  # [kmer_len, uniq_kh_arr, uniq_kh_cnt_arr]
    with open(kmer_cnt_file, "rb") as fh:
        res_list = pickle.load(fh)
    assert res_list[0] == kmer_len
    uniq_kh_arr, uniq_kh_cnt_arr = res_list[1], res_list[2]

    # ensure there are enough seqs to sample,
    sampling_flag = True
    if n_total_sample > sum(uniq_kh_cnt_arr):
        warnings.warn(f"The number of samples n_sample={n_total_sample} is larger than the original " +
                      f"data n_seq={sum(uniq_kh_cnt_arr)}, process and return original data.")
        sampling_flag = False

    # calculate hamming distance to each consensus
    n_conseq = len(conseq_list)
    n_uniq_kmer = len(uniq_kh_arr)
    ham_dist_mat = np.zeros((n_conseq, n_uniq_kmer), dtype=int)
    rc_flag_mat = np.zeros((n_conseq, n_uniq_kmer), dtype=bool) # if the min ham dist is from reverse complement
    for i, conseq in enumerate(conseq_list):
        conseq_kh = kmer2hash(conseq)
        dist_arr = cal_hamming_dist_head(uniq_kh_arr, conseq_kh, kmer_len, len(conseq))
        if revcom_mode:
            rc_conseq_kh = revcom_hash(conseq_kh, len(conseq))
            assert conseq_kh <= rc_conseq_kh
            rc_dist_arr = cal_hamming_dist_tail(uniq_kh_arr, rc_conseq_kh, kmer_len, len(conseq))  # revcom
            rc_flag_mat[i] = rc_dist_arr < dist_arr
            dist_arr = np.minimum(dist_arr, rc_dist_arr)
        ham_dist_mat[i] = dist_arr

    # label noise kmers for each consensus
    for i, conseq in enumerate(conseq_list):
        tmp_max_ham_dist = motif_def_dict[len(conseq)].max_ham_dist
        ham_dist_mat[i][ham_dist_mat[i] > tmp_max_ham_dist] = kmer_len  # maximum distance is kmer_len, so noise

    # assign label
    min_dist_arr = np.min(ham_dist_mat, axis=0)
    min_dist_ind_arr = np.argmin(ham_dist_mat, axis=0)
    min_dist_ind_arr[min_dist_arr > motif_def_dict[kmer_len].max_ham_dist] = n_conseq
    label_arr = min_dist_ind_arr

    # # assign label
    # min_dist_arr = np.min(ham_dist_mat, axis=0)
    # min_dist_ind_arr = np.argmin(ham_dist_mat, axis=0)
    # min_dist_ind_arr[min_dist_arr > motif_def_dict[kmer_len].max_ham_dist] = n_conseq
    # for i, conseq in enumerate(conseq_list):
    #     if len(conseq) < kmer_len:
    #         tmpinds = np.where(min_dist_ind_arr == i)[0]
    #         tmpinds = tmpinds[ min_dist_arr[tmpinds] > motif_def_dict[len(conseq)].max_ham_dist ]
    #         min_dist_ind_arr[tmpinds] = n_conseq
    # label_arr = min_dist_ind_arr # each conseq has a label, random seq corresponds to n_conseq

    # update kmer hash list for revcom mode, such that all the kmers in a hamming ball align with the consensus sequence
    if revcom_mode:
        for i, conseq in enumerate(conseq_list):
            tmpinds = np.where(label_arr == i)[0]
            for j in np.where(rc_flag_mat[i][tmpinds])[0]:
                tmpind = tmpinds[j]
                uniq_kh_arr[tmpind] = revcom_hash(uniq_kh_arr[tmpind], kmer_len)

    # no sampling case
    if not sampling_flag:
        return uniq_kh_arr, uniq_kh_cnt_arr, label_arr, conseq_list

    sample_cnt_arr = np.bincount(label_arr, weights=uniq_kh_cnt_arr)
    motif_weights = sample_cnt_arr[:-1] / sum(sample_cnt_arr[:-1])
    sample_cnt_arr[:-1] = np.around(n_motif_kmer * motif_weights)
    sample_cnt_arr[-1] = n_total_sample - sum(sample_cnt_arr[0:-1])
    sample_cnt_arr = sample_cnt_arr.astype(int)
    assert len(sample_cnt_arr) == n_conseq + 1

    # sampling
    samp_inds = []
    samp_cnts = []
    for c in range(n_conseq+1):
        c_inds = np.where(label_arr == c)[0]
        ws = uniq_kh_cnt_arr[c_inds]
        ws = ws/sum(ws)
        tmpcnts = np.random.multinomial(sample_cnt_arr[c], ws, size=1).squeeze()
        samp_inds.append(c_inds[tmpcnts > 0])
        samp_cnts.append(tmpcnts[tmpcnts > 0])

    samp_inds = np.concatenate(samp_inds)
    samp_cnts = np.concatenate(samp_cnts)
    samp_kh_arr = uniq_kh_arr[samp_inds]
    samp_label_arr = label_arr[samp_inds]

    return samp_kh_arr, samp_cnts, samp_label_arr, conseq_list # output pkl, txt file


def ex_hamball_kh_arr(res_dir: str, conseq: str, max_ham_dist: int=-1, motif_def_file: str=None, revcom_mode=True):
    """
    Extract kmer hash for all kmers of the Hamming ball centered on the input consensus sequence 
    Args:
        res_dir: result directory
        conseq: consensus sequence
        max_ham_dist: maximum Hamming distance of the Hamming ball
        motif_def_file: motif definition file
        revcom_mode: reverse complment mode
    Returns:
        tuple (unique kmer hash, unique kmer cnt)
    """
    conseq = conseq.upper()
    assert all([e in ("A", "C", "G", "T") for e in conseq])
    kmer_len = len(conseq)
    conseq_kh = kmer2hash(conseq)
    rc_conseq_kh = revcom_hash(conseq_kh, kmer_len)
    assert conseq_kh <= rc_conseq_kh

    assert Path(motif_def_file).exists()
    assert Path(res_dir).exists()
    res_path = Path(res_dir)

    # load kmer counts
    kmer_cnt_file = res_path / FileNameDict["kmer_count_dir"] / f"k{kmer_len}.pkl"  # [kmer_len, uniq_kh_arr, uniq_kh_cnt_arr]
    with open(kmer_cnt_file, "rb") as fh:
        res_list = pickle.load(fh)
    assert res_list[0] == kmer_len
    uniq_kh_arr, uniq_kh_cnt_arr = res_list[1], res_list[2]

    if max_ham_dist == -1:
        motif_def_dict = init_motif_def_dict(motif_def_file)
        max_ham_dist = motif_def_dict[kmer_len].max_ham_dist

    n_uniq_kmer = len(uniq_kh_arr)
    rc_flag_arr = np.zeros(n_uniq_kmer, dtype=bool)  # if the min ham dist is from reverse complement
    dist_arr = cal_hamming_dist(uniq_kh_arr, conseq_kh, kmer_len)
    if revcom_mode:
        rc_dist_arr = cal_hamming_dist(uniq_kh_arr, rc_conseq_kh, kmer_len)
        rc_flag_arr = rc_dist_arr < dist_arr
        dist_arr = np.minimum(dist_arr, rc_dist_arr)

    hamball_flag_arr = dist_arr <= max_ham_dist
    # update kmer hash list for revcom mode, such that all the kmers in a hamming ball align with the consensus sequence
    if revcom_mode:
        tmp_flag_arr = np.logical_and(rc_flag_arr, hamball_flag_arr)
        tmpinds = np.where(tmp_flag_arr)[0]
        for i in tmpinds:
            uniq_kh_arr[i] = revcom_hash(uniq_kh_arr[i], kmer_len)

    return uniq_kh_arr[hamball_flag_arr], uniq_kh_cnt_arr[hamball_flag_arr]


def cal_cnt_mat(uniq_kh_arr, uniq_kh_cnt_arr, kmer_len):
    cnt_mat = np.zeros((4, kmer_len), dtype=int)

    for kh, cnt in zip(uniq_kh_arr, uniq_kh_cnt_arr):
        kmer = hash2kmer(kh, kmer_len)
        kmer_arr = dna2arr(kmer, dtype=np.uint8, append_missing_val_flag=False)
        for i, b in enumerate(kmer_arr):
            cnt_mat[b][i] += cnt
    return cnt_mat


def _draw_logo(cnt_mat_numpy_file: str, output_fig_file=None):
    cntmat = np.loadtxt(cnt_mat_numpy_file, delimiter=",")
    cntmat = np.transpose(cntmat)
    n_pos = len(cntmat)
    cnt_df = pd.DataFrame(data=cntmat, index=np.arange(n_pos), columns=["A", "C", "G", "T"])
    cnt_df_info = logomaker.transform_matrix(cnt_df, from_type="counts", to_type="information")
    crp_logo = logomaker.Logo(cnt_df_info, font_name='Arial')
    if output_fig_file:
        plt.savefig(output_fig_file)



def gen_motif_co_occurence_mat(input_fasta_file: str, conseq_list: List[str], motif_def_dict: dict, min_motif_cnt=1, revcom_mode=True):
    """
    get the co-occurence count matrix of different motifs by scanning each read in the input fasta file
    Args:
        input_fasta_file: input fasta file
        conseq_list: consensus sequences of motifs
        motif_def_dict: motif definition table
        min_motif_cnt: minimum number of observed motif in each read
        revcom_mode: if reverse complement should be considered

    Returns:
        n_motif x n_motif count matrix

    """
    def read_stream(fh):
        for rec in SeqIO.parse(fh, "fasta"):
            yield str(rec.seq).upper()

    def read_fasta(input_fasta_file):
        if input_fasta_file.endswith(".gz"):
            with gzip.open(input_fasta_file, "rt") as fh:
                yield from read_stream(fh)
        else:
            with open(input_fasta_file, "r") as fh:
                yield from read_stream(fh)

    n_conseq = len(conseq_list)
    assert  n_conseq > 0
    res_mat = np.zeros((n_conseq, n_conseq), dtype=int)
    for seq in read_fasta(input_fasta_file):
        cnt_arr = get_motif_co_occurence_cnt_arr(seq, conseq_list, motif_def_dict, revcom_mode=revcom_mode)
        motif_inds = np.where(cnt_arr >= min_motif_cnt)[0]
        for i in range(len(motif_inds)):
            for j in range(i, len(motif_inds)):
                res_mat[motif_inds[i], motif_inds[j]] += 1
    diag_vec = np.diag(res_mat)
    diag_inds = np.diag_indices(n_conseq)
    res_mat += np.transpose(res_mat)
    res_mat[diag_inds] = diag_vec

    return res_mat


def get_motif_co_occurence_cnt_arr(dna_seq: str, conseq_list: List[str], motif_def_dict: dict, revcom_mode=True) -> np.ndarray:
    """
    get the counts of each input motif (hamming ball) for a given DNA sequence
    Args:
        dna_seq: input dna sequence
        conseq_list: consensus sequence list, consensus length can be different
        motif_def_dict: motif definition dictionary, kmer_len : MotifDef obj
        revcom_mode: if reverse complement should be considered
    Returns:
        a numpy array showing the counts for each motif
    """

    assert len(conseq_list) > 0
    conseq_kh_list = []
    conseq_len_list = []
    for conseq in conseq_list:
        conseq_len_list.append(len(conseq))
        conseq_kh_list.append(kmer2hash(conseq))

    seq_np_arr = dna2arr(dna_seq, append_missing_val_flag=False)
    kh_arr_dict = {}
    uniq_conseq_len_arr = np.unique(conseq_len_list)
    for kmer_len in uniq_conseq_len_arr:
        hash_arr = comp_kmer_hash_taichi(seq_np_arr, kmer_len)
        uniq_kh_arr, uniq_kh_cnt_arr = count_uniq_hash(hash_arr, kmer_len)
        kh_arr_dict[kmer_len] = (uniq_kh_arr, uniq_kh_cnt_arr)

    res = np.zeros(len(conseq_list), dtype=int)
    for i in range(len(conseq_list)):
        conseq_kh, conseq_len = conseq_kh_list[i], conseq_len_list[i]
        uniq_kh_arr, uniq_kh_cnt_arr = kh_arr_dict[conseq_len]
        max_ham_dist = motif_def_dict[conseq_len].max_ham_dist
        res[i] = _get_motif_cnt(uniq_kh_arr, uniq_kh_cnt_arr, conseq_len, conseq_kh, max_ham_dist,
                                rev_com_mode=revcom_mode)
    return res


def _get_motif_cnt(uniq_kh_arr: np.array, uniq_kh_cnt_arr: np.array,
                               kmer_len: int, conseq_kh: np.uint64,
                               max_ham_dist: int, rev_com_mode=True) -> int:
    """
    get the count of motif from input kh_arr
    Args:
        uniq_kh_arr: unique kmer hash array
        uniq_kh_cnt_arr: count of each unique kmer hash
        kmer_len: kmer length
        conseq_kh: hash of consequence sequence
        max_ham_dist: maximum hamming distance of hamming ball
        rev_com_mode: if revcom should be considered
    Returns:
        number of kmers fall in the hamming ball
    """
    dist_arr = cal_hamming_dist(uniq_kh_arr, conseq_kh, kmer_len)
    if rev_com_mode:
        rc_kh = revcom_hash(conseq_kh, kmer_len)
        rc_dist_arr = cal_hamming_dist(uniq_kh_arr, rc_kh, kmer_len)  # revcom
        dist_arr = np.minimum(dist_arr, rc_dist_arr)
    return int(np.sum(uniq_kh_cnt_arr[dist_arr <= max_ham_dist]))


import numpy as np
from Bio import SeqIO
import gzip
import click
from typing import Callable
import pickle
import taichi as ti
from typing import List, Tuple
from taichi.algorithms import parallel_sort
from scipy.stats import norm

from pathlib import Path
from dataclasses import dataclass, fields
import tomllib
import tomli_w

from .taichi_core import (kmer2hash_kernel_uint32, kmer2hash_kernel_uint64,
                         revcom_hash_kernel_uint32, revcom_hash_kernel_uint64,
                         cal_ham_dist_kernel_uint32, cal_ham_dist_kernel_uint64,
                         cal_partial_ham_dist_head_kernel_uint32, cal_partial_ham_dist_head_kernel_uint64,
                         cal_partial_ham_dist_tail_kernel_uint32, cal_partial_ham_dist_tail_kernel_uint64)

import pandas as pd

from importlib.resources import files, as_file

FileNameDict = {
    "default_config_file": "default_config.toml",
    "config_file": "config.toml",
    "default_motif_def_file": "default_motif_def_table.csv",
    "motif_def_file": "motif_def_table.csv",
    "processed_fasta_file": "input.bin.pkl",
    "kmer_count_dir": "kmer_count",
    "hamball_dir": "hamming_balls",
    "co_occur_mat_file": "co_occurence_mat.tsv",
    "co_occur_mat_norm_file": "co_occurence_mat.norm.tsv",
    "candidate_conseq_file": "candidate_conseq.csv",
    "final_conseq_file": "final_conseq.txt",
    "final_conseq_info_file": "final_conseq.info.txt",
    "sample_kmer_pkl_file": "sample_kmers.pkl",
    "sample_kmer_txt_file": "sample_kmers.tsv",
    "sample_kmer_hamdist_mat_file": "sample_kmer_hamdist_mat.pkl",
    "ld_data_file": "low_dim_data.tsv",
    "ld_fig_file_stem": "ld_data",
}


ti.set_logging_level(ti.ERROR)

MISSING_VAL = 255
GPU_MODE = False
ti.init(arch=ti.cuda, default_ip=ti.i64)
if ti.cfg.arch == ti.cuda:
    GPU_MODE = True
    print("GPU is available")
else:
    GPU_MODE = False
    print("GPU is not available")


@click.command(name="preproc")
@click.option(
    '--fasta_file',
    type=str,
    help='Input fasta file',
    required=True
    )
@click.option(
    '--res_dir',
    type=str,
    default=".",
    help='Result directory for storing all outputs',
    required=False
    )
@click.option(
    '--debug',
    type=bool,
    default=False,
    help='display debug information.',
    required=False
    )
def preproc(fasta_file: str, res_dir=".", debug=False):
    _preproc(fasta_file, res_dir, debug)


def read_default_config_file(debug=False):
    config_file_path = files(__package__).joinpath(FileNameDict["default_config_file"])  # may be in a zip file
    with as_file(config_file_path) as fh:
        with open(fh, "rb") as fh1:
            default_config_dict = tomllib.load(fh1)
    if debug:
        print(default_config_dict)
    return default_config_dict


def gen_motif_def_dict(config_dict: dict, debug=False):
    """
    generate the motif definition dict
    Args:
        config_dict: configuration dictionary derived from .toml file
        debug: if print debug information

    Returns:
        motif definition dictionary
    """
    motif_def_file = config_dict["motif_discovery"]["motif_def_file"]
    if motif_def_file == "default":
        motif_def_file_path = files(__package__).joinpath(FileNameDict["default_motif_def_file"])  # may be in a zip file
        with as_file(motif_def_file_path) as fh:
            motif_def_dict = init_motif_def_dict(fh, p_value_cutoff=config_dict["motif_discovery"]["p_value_cutoff"])
    else:
        assert Path(motif_def_file).exists()
        motif_def_dict = init_motif_def_dict(motif_def_file, p_value_cutoff=config_dict["motif_discovery"]["p_value_cutoff"])

    if debug:
        print(motif_def_dict)

    return motif_def_dict


def _preproc(fasta_file: str, res_dir=".", debug=False):
    input_fasta_file = fasta_file
    assert Path(input_fasta_file).exists()
    if not Path(res_dir).exists():
        Path(res_dir).mkdir()

    # read configuration file
    config_file_name = FileNameDict["config_file"]
    config_file_path = Path(res_dir) / config_file_name
    if Path(config_file_path).exists():
        with open(config_file_path, "rb") as fh:
            config_dict = tomllib.load(fh)
    else:
        config_dict = read_default_config_file(debug=debug)

    # save configuration file in output directory
    if not Path(config_file_path).exists() or config_dict["general"].get("input_fasta_file") is None:
        config_dict["general"]["input_fasta_file"] = input_fasta_file
        config_dict["general"]["res_dir"] = res_dir
        with open(config_file_path, 'wb') as fh:
            tomli_w.dump(config_dict, fh)

    # read motif definition file
    motif_def_dict = gen_motif_def_dict(config_dict, debug=debug)
    kmer_len_list = sorted([e for e in motif_def_dict if isinstance(e, int)])
    headers = motif_def_dict[kmer_len_list[0]].get_field_names()

    # save motif definition table in result directory
    motif_def_file = FileNameDict["motif_def_file"]
    motif_def_file_path = Path(res_dir) / motif_def_file
    with open(motif_def_file_path, "w+") as fh:
        fh.write(headers + "\n")
        for kmer_len in kmer_len_list:
            fh.write(str(motif_def_dict[kmer_len]) + "\n")

    # process input fasta file
    proc_input(config_dict["general"]["input_fasta_file"], config_dict["general"]["res_dir"],
               out_bin_file_name=FileNameDict["processed_fasta_file"], debug=debug)
    return config_dict, motif_def_dict


def proc_input(input_fasta_file: str, res_dir=".", out_bin_file_name: str="input.bin.pkl", debug=True):
    """
    process input fasta file, convert fasta file to a binary file "input.bin.pkl" in the output directory
    Args:
        input_fasta_file: input fasta file
        res_dir: output directory
        out_bin_file_name: name of the output binary file
        debug: if displaying debug information
    Returns: None
    """
    def get_file_size(fasta_file: str):
        file_size = 0
        for arr in read_dnaseq_file(fasta_file):
            file_size = file_size + len(arr)
        return file_size

    assert Path(input_fasta_file).exists()
    assert Path(res_dir).exists()
    assert out_bin_file_name.endswith(".pkl")

    buffer_size = get_file_size(input_fasta_file)
    #input_binary_file = os.path.join(res_dir, out_bin_file_name)
    input_binary_file = str(Path(res_dir) / out_bin_file_name)

    if debug:
        print(f"Convert input file={input_fasta_file} into binary file {input_binary_file}. buffer_size={buffer_size/2**30}GB.")

    buffer = Buffer(buffer_size)
    # convert input fasta file into binary file
    convert_fasta_to_binary(input_fasta_file, buffer, input_binary_file)

    print(f"input binary file {input_binary_file} generated.\n")


@dataclass
class MotifDef:
    kmer_len: int
    p_uniform: float
    max_ham_dist: int
    ratio_mu: float
    ratio_std: float
    ratio_cutoff: float

    @classmethod
    def get_field_names(cls):
        return ",".join(field.name for field in fields(cls))

    def __str__(self):
        return ",".join(str(getattr(self, field.name)) for field in fields(self))


def arr2dna(dna_np_arr: np.ndarray) -> str:
    base_map = {0: 'A', 1: 'C', 2: 'G', 3: 'T', MISSING_VAL: 'N'}
    res = [base_map[i] for i in dna_np_arr]
    return "".join(res)


def dna2arr(dna_str, dtype=np.uint8, append_missing_val_flag=True) -> np.ndarray:
    """
    convert an input DNA string to numpy uint8 array, with a missing value appended
    Args:
        dna_str: a DNA sequence, all letters should be upper case
        dtype: data type for storing DNA string
        append_missing_val_flag: if a missing value is appended to the end of the array
    Returns:
        a numpy array
    """
    if append_missing_val_flag:
        res = np.empty(len(dna_str) + 1, dtype=dtype)
    else:
        res = np.empty(len(dna_str), dtype=dtype)
    base_map = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
    for i, b in enumerate(dna_str):
        res[i] = base_map.get(b, MISSING_VAL)
    if append_missing_val_flag:
        res[-1] = MISSING_VAL  # add a separator to the end of the string
    return res


def reverse_complement(seq):
    complement = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
    return ''.join(complement[base] for base in reversed(seq))


class Buffer:
    def __init__(self, buffer_size: int = 2 ** 26, dtype=np.uint8, id=None):
        # 2**26 bytes are 64MB, 2**30 bytes are 1GB
        self.buffer_size = buffer_size
        self.dtype = dtype
        self.buffer = np.full(buffer_size, MISSING_VAL, dtype=dtype)
        self.pointer = 0
        self.data_to_write = None
        self.is_full = False
        self.id = None

    def append(self, data: np.ndarray):
        # append input data into the buffer, return flag of success
        assert data.dtype == self.buffer.dtype
        data_size = len(data)
        buffer_space = self.buffer_size - self.pointer

        # Check if the buffer has enough space to accommodate the new data
        if data_size > buffer_space:
            self.data_to_write = data
            self.is_full = True
            return False
        else:
            # Append the data to the buffer
            self.buffer[self.pointer: (self.pointer + data_size)] = data
            self.pointer += data_size
            return True

    def flush(self):
        # flush the buffer such that it can be used again, data_to_write needs to be handled first
        assert self.data_to_write is None
        self.pointer = 0
        self.data_to_write = None
        self.is_full = False
        self.id = None


def read_dnaseq_file(file_name, file_type="fasta") -> np.ndarray:
    """
    file_name: input DNA sequence file name
    file_type: fasta, fastq,
    """

    def read_stream(fh):
        for rec in SeqIO.parse(fh, file_type):
            yield dna2arr(str(rec.seq).upper(), append_missing_val_flag=True)

    if file_name.endswith(".gz"):
        with gzip.open(file_name, "rt") as fh:
            yield from read_stream(fh)
    else:
        with open(file_name, "r") as fh:
            yield from read_stream(fh)


def convert_fasta_to_binary(fasta_file: str | Path, buffer: Buffer, out_pkl_file: str | Path) -> None:
    def write_numpy_array(np_arr: np.ndarray, file_name: str | Path):
        assert str(file_name).endswith(".pkl")
        with open(file_name, "wb") as fh:
            pickle.dump(np_arr, fh)

    for arr in read_dnaseq_file(fasta_file):
        flag = buffer.append(arr)
        assert flag
    assert not buffer.is_full
    write_numpy_array(buffer.buffer[0:buffer.pointer], out_pkl_file)


# get the numpy dtype to store the kmer cnts
def get_cnt_dtype(kmer_len: int) -> np.dtype:
    if kmer_len < 16:
        return np.int32
    else:
        return np.int64


# get the hash dtype for given kmer length
def get_hash_dtype(kmer_len) -> np.dtype:
    if 0 < kmer_len < 16:
        return np.uint32
    elif kmer_len < 32:
        return np.uint64
    else:
        raise Exception(f"max_kmer_len=31, kmer_len={kmer_len} is greater the maximum value.")


# get the hash value for invalid kmers
def get_invalid_hash(dtype: Callable[[int], np.dtype]):
    return dtype(np.iinfo(dtype).max)


def my_parallel_sort(arr: np.ndarray):
    if arr.dtype == np.uint32:
        dtype = ti.u32
    elif arr.dtype == np.uint64:
        dtype = ti.u64
    else:
        raise Exception(f"unknown arr.dtype={arr.dtype}")

    tarr = ti.field(dtype=dtype, shape=len(arr))
    tarr.from_numpy(arr)
    parallel_sort(tarr)
    arr[:] = tarr.to_numpy()


def my_unique(hash_arr, hash_dtype, invalid_hash, cnt_dtype):
    my_parallel_sort(hash_arr)
    curr_hash = invalid_hash
    n_uniq_val = 0
    for i in range(len(hash_arr)):
        if hash_arr[i] == invalid_hash:
            break
        elif hash_arr[i] == curr_hash:
            continue
        else:
            curr_hash = hash_arr[i]
            n_uniq_val += 1
    unique_hash = np.empty(n_uniq_val, dtype=hash_dtype)
    counts = np.zeros(n_uniq_val, dtype=cnt_dtype)
    curr_hash = invalid_hash
    i_uniq = -1
    for i in range(len(hash_arr)):
        if hash_arr[i] == invalid_hash:
            break
        elif hash_arr[i] == curr_hash:
            counts[i_uniq] += 1
        else:
            curr_hash = hash_arr[i]
            i_uniq += 1
            unique_hash[i_uniq] = hash_arr[i]
            counts[i_uniq] += 1
    return unique_hash, counts


def kmer2hash(kmer: str) -> np.uint64:
    """
    kmer: input sequence
    return: a hash code
    """
    k = len(kmer)
    assert k < 32, "kmer should be shorted than 32 bases"
    base_map = {'A': 0, 'C': 1, 'G': 2, 'T': 3}
    base = {bk: np.uint64(base_map[bk]) for bk in base_map}

    kh = base[kmer[0]]
    for tb in kmer[1:]:
        kh = kh << np.uint64(2)
        kh += base[tb]
    return kh


def hash2kmer(hashkey: np.uint64, k: int) -> str:
    """
    hashkey: hash key of kmer, numpy
    k: length of kmer
    """
    hashkey = np.uint64(hashkey)
    base = np.array('ACGT', 'c')
    arr = np.chararray(k)
    mask = np.uint64(3) # 0b11
    arr[-1] = base[ mask & hashkey]
    for i in range(2,k+1):
        hashkey = (hashkey >> np.uint64(2))
        arr[-i] = base[mask & hashkey]
    return arr.tobytes().decode("utf-8")


def comp_kmer_hash_taichi(seq_np_arr: np.ndarray, kmer_len: int) -> np.ndarray:
    """
    Compute kmer hash for each kmer from the input seq array
    Args:
        seq_np_arr: a numpy array that contain DNA sequences, A-0, C-1,
        kmer_len: length of kmer
    Returns: Tuple (uniq_kh_arr, uniq_kh_cnt_arr)
    """

    missing_val = np.uint8(MISSING_VAL)
    missing_val_arr = np.array([missing_val])
    seq_len = len(seq_np_arr)

    hash_dtype = get_hash_dtype(kmer_len)
    invalid_hash = get_invalid_hash(hash_dtype)
    invalid_hash_arr = np.array([invalid_hash])

    hash_arr = np.empty(seq_len, dtype=hash_dtype)
    if hash_dtype == np.uint32:
        kmer2hash_kernel_uint32(seq_np_arr, seq_len, kmer_len, hash_arr, invalid_hash_arr, missing_val_arr)
    elif hash_dtype == np.uint64:
        kmer2hash_kernel_uint64(seq_np_arr, seq_len, kmer_len, hash_arr, invalid_hash_arr, missing_val_arr)
    else:
        raise Exception(f"Unknown kmer hash type hash_dtype={hash_dtype}")
    return hash_arr


def count_uniq_hash(hash_arr: np.ndarray, kmer_len):
    hash_dtype = get_hash_dtype(kmer_len)
    invalid_hash = get_invalid_hash(hash_dtype)
    cnt_arr_dtype = get_cnt_dtype(kmer_len)

    if GPU_MODE:
        unique_hash, counts = my_unique(hash_arr, hash_dtype, invalid_hash, cnt_arr_dtype)
    else:
        unique_hash, counts = np.unique(hash_arr, return_counts=True)
        inds = unique_hash != invalid_hash
        unique_hash = unique_hash[inds]
        counts = counts[inds].astype(cnt_arr_dtype)
    # hash_counts_dict = Counter(dict(zip(unique_hash, counts)))

    # return hash_counts_dict
    return unique_hash, counts


def cal_hamming_dist(kh_arr: np.ndarray, consensus_kh: np.uint64, kmer_len: int) -> np.ndarray:
    """
    calculate the Hamming distances between each element in kh_arr and the consensus sequence
    Args:
        kh_arr: kmer hash array
        consensus_kh: kmer hash of the consensus sequence
        kmer_len: kmer length
    Returns:
        Hamming distance array, np.ndarray object
    """
    ham_dist_arr = np.empty_like(kh_arr, dtype=np.uint8)
    hash_dtype = get_hash_dtype(kmer_len)
    consensus_kh_arr = np.array([consensus_kh], dtype=hash_dtype)
    hash_arr_size = len(kh_arr)

    if hash_dtype == np.uint32:
        cal_ham_dist_kernel_uint32(kh_arr, consensus_kh_arr, ham_dist_arr, hash_arr_size, kmer_len)
    elif hash_dtype == np.uint64:
        cal_ham_dist_kernel_uint64(kh_arr, consensus_kh_arr, ham_dist_arr, hash_arr_size, kmer_len)
    else:
        raise Exception(f"Unknown kmer hash type hash_dtype={hash_dtype}")
    return ham_dist_arr


def cal_hamming_dist_head(kh_arr: np.ndarray, consensus_kh: np.uint64, kmer_len: int, consensus_len: int) -> np.ndarray:
    """
    calculate the Hamming distances between each element in kh_arr and the consensus sequence
    The consensus sequence might be shorter than the kmer length
    Only match the consensus from the head of each kmer
    Args:
        kh_arr: kmer hash array
        consensus_kh: kmer hash of the consensus sequence
        kmer_len: kmer length
        consensus_len: length of the consensus sequence
    Returns:
        Hamming distance array, np.ndarray object
    """
    assert consensus_len <= kmer_len

    ham_dist_arr = np.empty_like(kh_arr, dtype=np.uint8)
    hash_dtype = get_hash_dtype(kmer_len)
    consensus_kh_arr = np.array([consensus_kh], dtype=hash_dtype)
    hash_arr_size = len(kh_arr)

    if hash_dtype == np.uint32:
        cal_partial_ham_dist_head_kernel_uint32(kh_arr, consensus_kh_arr, ham_dist_arr, hash_arr_size, kmer_len,
                                                consensus_len)
    elif hash_dtype == np.uint64:
        cal_partial_ham_dist_head_kernel_uint64(kh_arr, consensus_kh_arr, ham_dist_arr, hash_arr_size, kmer_len,
                                                consensus_len)
    else:
        raise Exception(f"Unknown kmer hash type hash_dtype={hash_dtype}")
    return ham_dist_arr


def cal_hamming_dist_tail(kh_arr: np.ndarray, consensus_kh: np.uint64, kmer_len: int, consensus_len: int) -> np.ndarray:
    """
    calculate the Hamming distances between each element in kh_arr and the consensus sequence
    The consensus sequence might be shorter than the kmer length
    Only match the consensus from the tail of each kmer
    Args:
        kh_arr: kmer hash array
        consensus_kh: kmer hash of the consensus sequence
        kmer_len: kmer length
        consensus_len: length of the consensus sequence
    Returns:
        Hamming distance array, np.ndarray object
    """
    assert consensus_len <= kmer_len

    ham_dist_arr = np.empty_like(kh_arr, dtype=np.uint8)
    hash_dtype = get_hash_dtype(kmer_len)
    consensus_kh_arr = np.array([consensus_kh], dtype=hash_dtype)
    hash_arr_size = len(kh_arr)

    if hash_dtype == np.uint32:
        cal_partial_ham_dist_tail_kernel_uint32(kh_arr, consensus_kh_arr, ham_dist_arr, hash_arr_size, kmer_len,
                                                consensus_len)
    elif hash_dtype == np.uint64:
        cal_partial_ham_dist_tail_kernel_uint64(kh_arr, consensus_kh_arr, ham_dist_arr, hash_arr_size, kmer_len,
                                                consensus_len)
    else:
        raise Exception(f"Unknown kmer hash type hash_dtype={hash_dtype}")
    return ham_dist_arr


def mask_input(seq_np_arr: np.ndarray, kmer_len: int, consensus_kh_arr: np.ndarray, max_hamball_dist_arr: np.ndarray):
    """
    mask the input seq_np_arr such that occurences of all kmers in given consensus hamming ball as masked
    consensus_kh_arr can contain revcom, repetitive kmers
    Args:
        seq_np_arr: converted numpy arr of the input DNA sequence
        kmer_len: kmer length
        consensus_kh_arr: consensus hash arr, all consensus must be the same length
        max_hamball_dist_arr: maximum Hamming distance in the Hamming ball for the corresponding consensus hash
    Returns:
        seq_np_arr with relevant positions in seq_np_arr overwritten as Missing value
    """
    def _mask_one_hamball(seq_np_arr, kh_hash_arr, consensus_kh, max_hamball_dist, kmer_len):
        ham_dist_arr = cal_hamming_dist(kh_hash_arr, consensus_kh, kmer_len)
        flag_arr = ham_dist_arr <= max_hamball_dist
        for i, flag in enumerate(flag_arr):
            if flag:
                j = i + kmer_len if i + kmer_len < len(seq_np_arr) else len(seq_np_arr)
                seq_np_arr[i:j] = MISSING_VAL
        del ham_dist_arr

    kh_hash_arr = comp_kmer_hash_taichi(seq_np_arr, kmer_len)
    for consensus_kh, max_hamball_dist in zip(consensus_kh_arr, max_hamball_dist_arr):
        _mask_one_hamball(seq_np_arr, kh_hash_arr, consensus_kh, max_hamball_dist, kmer_len)

    del kh_hash_arr
    return seq_np_arr


def get_revcom_hash_arr(in_hash_arr: np.ndarray, kmer_len: int) -> np.ndarray:
    hash_dtype = get_hash_dtype(kmer_len)
    mask_arr = np.array([(1 << 2 * kmer_len) - 1, 3], dtype=hash_dtype) # mask and twobit_mask

    out_hash_arr = np.empty_like(in_hash_arr)
    hash_arr_size = len(in_hash_arr)
    if hash_dtype == np.uint32:
        revcom_hash_kernel_uint32(in_hash_arr, out_hash_arr, mask_arr, kmer_len, hash_arr_size)
    elif hash_dtype == np.uint64:
        revcom_hash_kernel_uint64(in_hash_arr, out_hash_arr, mask_arr, kmer_len, hash_arr_size)
    return out_hash_arr


def revcom_hash(in_hash: np.uint64, kmer_len: int) -> np.uint64 | np.uint32:
    hash_dtype = get_hash_dtype(kmer_len)
    in_hash = hash_dtype(in_hash)
    mask_arr = np.array([(1 << 2 * kmer_len) - 1, 3], dtype=hash_dtype) # mask and twobit_mask
    mask = mask_arr[0]
    twobit_mask = mask_arr[1]

    com_hash = mask - in_hash  # complement hash
    # reverse
    ret_hash = twobit_mask & com_hash
    for i in range(kmer_len - 1):
        ret_hash = ret_hash << hash_dtype(2)
        com_hash = com_hash >> hash_dtype(2)
        ret_hash += twobit_mask & com_hash
    return ret_hash


def merge_revcom(uniq_kmer_hash_arr: np.ndarray, uniq_kh_cnt_arr: np.ndarray,
                 kmer_len: int, keep_lower_hash_flag=True) -> Tuple:
    """
    merge reverse complements by summing the counts of a pair.
    Only keep the smaller kmer hash for a pair of revcoms. Palindrome's counts keep the same.
    Args:
        uniq_kmer_hash_arr: unique kmer hash array sorted in ascending order
        uniq_kh_cnt_arr: counts of the corresponding kmer hash in uniq_kmer_hash_arr
        kmer_len: kmer length
        keep_lower_hash_flag: if keeping the lower hash as the key when merging a pair of reverse complements
    Returns:
        a tuple (kh_arr, kh_cnt_arr) with revcom merged
    """
    revcom_uniq_kmer_hash_arr = get_revcom_hash_arr(uniq_kmer_hash_arr, kmer_len)

    # merge kh counts with its revcom, both comm_kh_nat_inds and comm_kh_rc_inds should be the same set, but in different order
    comm_kh, comm_kh_nat_inds, comm_kh_rc_inds = np.intersect1d(uniq_kmer_hash_arr, revcom_uniq_kmer_hash_arr,
                                                                return_indices=True)
    uniq_kh_cnt_arr[comm_kh_nat_inds] += uniq_kh_cnt_arr[comm_kh_rc_inds]

    # handle palindromes, only count once for palindrome
    palindrome_inds = np.where(uniq_kmer_hash_arr == revcom_uniq_kmer_hash_arr)[0]
    uniq_kh_cnt_arr[palindrome_inds] = uniq_kh_cnt_arr[palindrome_inds] / 2

    # remove kmers that are revcom
    if keep_lower_hash_flag:
        inds = uniq_kmer_hash_arr[comm_kh_nat_inds] > revcom_uniq_kmer_hash_arr[comm_kh_nat_inds]
    else:
        inds = uniq_kmer_hash_arr[comm_kh_nat_inds] < revcom_uniq_kmer_hash_arr[comm_kh_nat_inds]

    other_inds = comm_kh_nat_inds[inds]
    uniq_kmer_hash_arr = np.delete(uniq_kmer_hash_arr, other_inds)
    revcom_uniq_kmer_hash_arr = np.delete(revcom_uniq_kmer_hash_arr, other_inds)
    uniq_kh_cnt_arr = np.delete(uniq_kh_cnt_arr, other_inds)

    # replace kmers that has a high/low hash in the list with low/high hash depending on keep_lower_hash_flag
    if keep_lower_hash_flag:
        inds = uniq_kmer_hash_arr > revcom_uniq_kmer_hash_arr
    else:
        inds = uniq_kmer_hash_arr < revcom_uniq_kmer_hash_arr
    uniq_kmer_hash_arr[inds] = revcom_uniq_kmer_hash_arr[inds]

    return uniq_kmer_hash_arr, uniq_kh_cnt_arr


def mask_ham_ball(seq_np_arr: np.ndarray, motif_def_dict: dict,
                    consensus_seq_list: List[str], max_ham_dist_list: List[int] = ()) -> np.ndarray:
    """
    mosk consensus's hamming ball given by the user
    Args:
        seq_np_arr: numpy array of the input sequence
        motif_def_dict: motif definition dictionary, kmer_len : MotifDef obj
        consensus_seq_list: consensus sequence list provided by the user
        max_ham_dist_list: maximum hamming distance for each consensus sequence
    Returns:
        masked sequence array
    """

    len_list = np.array([len(conseq) for conseq in consensus_seq_list])
    if len(max_ham_dist_list) == 0:
        # user only provide consensus sequence, but not max_ham_dist, we derive from motif_def_table
        max_ham_dist_list = [motif_def_dict[conseq_len].max_ham_dist for conseq_len in len_list]
    assert len(max_ham_dist_list) == len(consensus_seq_list)

    uniq_len_arr = np.unique(len_list)
    conseq_kh_arr_by_len = []
    conseq_max_hamdist_by_len = []
    for uniq_len in uniq_len_arr:
        inds = np.where(len_list == uniq_len)[0]
        tmp_list = [consensus_seq_list[i] for i in inds]
        tmp_kh_list = np.array([kmer2hash(seq) for seq in tmp_list])
        tmp_len_arr = np.array([max_ham_dist_list[i] for i in inds])
        conseq_kh_arr_by_len.append(tmp_kh_list)
        conseq_max_hamdist_by_len.append(tmp_len_arr)

    for uniq_len, conseq_kh_arr, hamdist_arr \
            in zip(uniq_len_arr, conseq_kh_arr_by_len, conseq_max_hamdist_by_len):
        kmer_len = uniq_len
        seq_np_arr = mask_input(seq_np_arr, kmer_len, conseq_kh_arr, hamdist_arr)

    return seq_np_arr


def init_motif_def_dict(motif_def_file, p_value_cutoff=1e-10) -> dict:
    motif_def_dict = {}
    motif_def_dict["p_value_cutoff"] = p_value_cutoff

    df = pd.read_csv(motif_def_file)
    for ind, row in df.iterrows():
        kmer_len = int(row['kmer_len'])
        p_uniform = row['p_uniform']
        max_ham_dist = int(row['max_ham_dist'])
        ratio_mu = row["ratio_mu"]
        ratio_std = row["ratio_std"]
        ratio_cutoff = norm.ppf(1-p_value_cutoff, loc=ratio_mu, scale=ratio_std)
        motif_def_dict[kmer_len] = MotifDef(kmer_len, p_uniform, max_ham_dist, ratio_mu, ratio_std, ratio_cutoff)

    return motif_def_dict


if __name__ == "__main__":
    #test_process_input()
    #test_comp_kmer_hash_taichi()
    #test_count_uniq_hash()
    #test_mask_input()
    #test_merge_revcom()
    #test_hash2kmer()
    #test_kmer2hash()

    #test_find_motif()
    #test_init_motif_def_dict()
    #test_mask_ham_ball()
    pass

    # import argparse
    #
    # parser = argparse.ArgumentParser(description='Process fasta files.')
    # parser.add_argument('--input_fasta_file', type=str, help='Path to the directory containing fasta files')
    # parser.add_argument("--min_k", type=int, help="Minimum kmer length.")
    # parser.add_argument("--max_k", type=int, help="Maximum kmer length.")
    # parser.add_argument("--preproc", action="store_true", help="Enable preprocessing.")
    # parser.add_argument("--net_flag", action="store_false", help="Enable preprocessing.")
    # parser.add_argument("-f", "--user_motif_file", default="", help="Path to the .txt file containing DNA sequences.")
    # args = parser.parse_args()
    # u_motif = read_user_motif_file(args.user_motif_file)
    #
    #
    # fasta_file = "../../data/raw/test.fasta"
    # min_k = 5
    # max_k = 10
    # u = ['AAAAA', 'AAAAAA', 'AAAAAAA']
    # preproc = True
    # main(args.path, u_motif, args.min_k, args.max_k, args.preproc)
    # #main(fasta_file, u, min_k, max_k, preproc)
